import argparse
from collections.abc import Mapping
import dataclasses
import pathlib
from typing import TYPE_CHECKING, Any

import jax.numpy as jnp
import jax.typing

if TYPE_CHECKING:
    import reax


def convert_params(params: dict[str, Any] | argparse.Namespace | None) -> dict[str, Any]:
    """Ensure parameters are a dict or convert to dict if necessary.

    Args:
        params (dict[str, Any] | argparse.Namespace | None): Object to
            be converted to `dict`.
    """
    # in case converting from namespace
    if isinstance(params, argparse.Namespace):
        params = vars(params)

    if params is None:
        params = {}

    return params


def sanitize_params(params: dict[str, Any]) -> dict[str, Any]:
    """Sanitize params."""
    for key in params:
        # JAX/numpy scalars to python base types
        if isinstance(params[key], (jnp.bool_, jnp.integer, jnp.floating)):
            params[key] = params[key].item()
        elif type(params[key]) not in [bool, int, float, str, jax.typing.ArrayLike]:
            params[key] = str(params[key])

    return params


def flatten_dict(
    params: Mapping[Any, Any], delimiter: str = "/", parent_key: str = ""
) -> dict[str, Any]:
    """Flatten hierarchical dict, e.g. ``{'a': {'b': 'c'}} -> {'a/b': 'c'}``.

    Args:
        parent_key (str, optional): defaults to "".
        params (Mapping[Any, Any]): The mapping containing
            hyperparameters.
        delimiter (str, optional): The delimiter to express hierarchy,
            defaults to "/".

    Examples:

        >>> flatten_dict({'a': {'b': 'c'}})
        {'a/b': 'c'}
        >>> flatten_dict({'a': {'b': 123}})
        {'a/b': 123}
        >>> flatten_dict({5: {'a': 123}})
        {'5/a': 123}
    """
    result: dict[str, Any] = {}
    for key, val in params.items():
        new_key = parent_key + delimiter + str(key) if parent_key else str(key)
        if dataclasses.is_dataclass(val):
            val = dataclasses.asdict(val)
        elif isinstance(val, argparse.Namespace):
            val = vars(val)

        if isinstance(val, Mapping):
            result = {**result, **flatten_dict(val, parent_key=new_key, delimiter=delimiter)}
        else:
            result[new_key] = val

    return result


def add_prefix(
    metrics: Mapping[str, jax.typing.ArrayLike | float], prefix: str, separator: str
) -> Mapping[str, jax.typing.ArrayLike | float]:
    """Insert prefix before each key in a dict, separated by the separator.

    Args:
        metrics (Mapping[str, jax.typing.ArrayLike | float]): Dictionary
            with metric names as keys and measured quantities as values.
        prefix (str): Prefix to insert before each key.
        separator (str): Separates prefix and original key name.
    :return s: Dictionary with prefix and separator inserted before each key.
    :rtype s: Mapping[str, jax.typing.ArrayLike | float]
    """
    if not prefix:
        return metrics

    return {f"{prefix}{separator}{key}": val for key, val in metrics.items()}


def scan_checkpoints(
    checkpoint_listener: "reax.listeners.ModelCheckpoint", logged_model_time: dict
) -> list[tuple[float, str, float, str]]:
    """Return the checkpoints to be logged.

    Args:
        checkpoint_listener ("reax.listeners.ModelCheckpoint"):
            Checkpoint listener reference.
        logged_model_time (dict): Dictionary containing the logged model
            times.
    """
    # get checkpoints to be saved with associated score
    checkpoints = {}
    if hasattr(checkpoint_listener, "last_model_path") and hasattr(
        checkpoint_listener, "current_score"
    ):
        checkpoints[checkpoint_listener.last_model_path] = (
            checkpoint_listener.current_score,
            "latest",
        )

    if hasattr(checkpoint_listener, "best_model_path") and hasattr(
        checkpoint_listener, "best_model_score"
    ):
        checkpoints[checkpoint_listener.best_model_path] = (
            checkpoint_listener.best_model_score,
            "best",
        )

    if hasattr(checkpoint_listener, "best_k_models"):
        for key, value in checkpoint_listener.best_k_models.items():
            checkpoints[key] = (value, "best_k")

    checkpoints = sorted(
        (pathlib.Path(path).stat().st_mtime, path, s, tag)
        for path, (s, tag) in checkpoints.items()
        if pathlib.Path(path).is_file()
    )
    checkpoints = [
        ckpt
        for ckpt in checkpoints
        if ckpt[1] not in logged_model_time or logged_model_time[ckpt[1]] < ckpt[0]
    ]
    return checkpoints



import abc
from collections.abc import Callable
import functools
from typing import ClassVar, Generic, TypeVar, cast

import equinox
from typing_extensions import override

__all__ = ("Metric", "FromFun")

_OutT = TypeVar("_OutT")


class Metric(equinox.Module, Generic[_OutT], metaclass=abc.ABCMeta):
    """The base class for all metrics.

    To be compatible with JAX Metrics are designed to be immutable meaning that when we 'update'
    them, the update will return a new instance with the results of the update.

    To create a new instance, use the ``.create()`` method with the data the metric will be
    calculated on.

    Updating can be done either with the ``.update()`` call when updating from raw data, or the
    ``.merge()`` call when merging an existing metric instance.
    """

    @classmethod
    def from_fun(cls, function: Callable, name: str = None) -> type["FromFun[_OutT]"]:
        """Create a new metric from this one where a function is called before passing it on to this
        metric.

        Args:
            cls
            function (Callable): The function to call.

        Returns:
            type["FromFun[OutT]"]: The new metric type.
        """

        # Construct the desired new name (e.g., 'AccuracyMetricFromFun')
        new_name = name or f"{cls.__name__}FromFun"

        class FromFunction(FromFun):
            metric = cls

            @classmethod
            def func(cls, *args, **kwargs) -> "Metric[_OutT]":
                """Fun function."""
                return function(*args, **kwargs)

        # Set the __name__ and __qualname__ attributes on the new class
        FromFunction.__name__ = new_name

        # __qualname__ often includes the module and containing class,
        # so we derive it from the base class's qualname
        FromFunction.__qualname__ = f"{cls.__qualname__}.{new_name}"

        return FromFunction

    def empty(self) -> "Metric[_OutT]":
        """Create a new empty instance.

        By default, this will call the constructor with no arguments, if needed, subclasses can
        overwrite this with custom behaviour.
        """
        return type(self)()

    @abc.abstractmethod
    def create(self, *args, **kwargs) -> "Metric[_OutT]":
        """Create a new metric instance from data."""

    def update(self, *args, **kwargs) -> "Metric[_OutT]":
        """Update the metric from new data and return a new instance."""
        return self.merge(self.create(*args, **kwargs))

    @abc.abstractmethod
    def merge(self, other: "Metric") -> "Metric[_OutT]":
        """Merge the metric with data from another metric instance of the same type."""

    @abc.abstractmethod
    def compute(self) -> _OutT:
        """Compute the metric."""


def hybrid_method(func: Callable):
    """
    A decorator that binds the function to the instance if called on an instance,
    or the class if called on the class.

    Implemented as a function returning a descriptor object.
    """

    class HybridDescriptor:
        def __get__(self, obj, objtype=None):
            if obj is None:
                # Called on the class -> bind to class
                return func.__get__(objtype, objtype)
            # Called on an instance -> bind to instance
            return func.__get__(obj, objtype)

    # Create the descriptor instance
    descriptor = HybridDescriptor()

    # Copy metadata (name, docstring) from the original function to the descriptor
    functools.update_wrapper(descriptor, func)

    return descriptor


class FromFun(Metric[_OutT]):
    """Helper class apply a function before passing the result to an existing metric."""

    metric: ClassVar[type[Metric[_OutT]]]
    func: ClassVar[Callable]
    _state: Metric[_OutT]

    def __init__(self, *args, state: Metric[_OutT] = None, **kwargs):
        super().__init__()
        if self.metric is None:
            raise RuntimeError(
                "The metric used by from_fun has not been set. "
                "This should be set as a class variable"
            )
        if state is None:
            state = self.metric(*args, **kwargs)
        self._state = state

    @property
    def is_empty(self) -> bool:
        return self._state is None

    @override
    @hybrid_method
    def empty(  # pylint: disable=no-self-argument, arguments-renamed
        self_or_cls,
    ) -> "FromFun[_OutT]":
        """Empty function."""
        if isinstance(self_or_cls, type):
            cls: "type[FromFun[_OutT]]" = cast(type(FromFun), self_or_cls)
            return cls(state=cls.metric.empty())  # pylint: disable=not-callable

        self = cast(FromFun, self_or_cls)
        return type(self)(state=self._state.empty())  # pylint: disable=protected-access

    @override
    @hybrid_method
    def create(  # pylint: disable=no-self-argument, arguments-differ
        self_or_cls, *args, **kwargs
    ) -> "FromFun":
        """Create function."""
        val = self_or_cls._call_fn(*args, **kwargs)

        if isinstance(self_or_cls, type):
            cls = self_or_cls
            metric = cls.metric
        else:
            cls = type(self_or_cls)
            metric = self_or_cls._state

        cls = cast(type(FromFun), cls)
        return cls(state=metric.create(*val))  # pylint: disable=not-callable

    @override
    def merge(self, other: "FromFun") -> "FromFun":
        """Merge function."""
        if self.is_empty:
            return other
        if other.is_empty:
            return self

        return type(self)(state=self._state.merge(other._state))  # pylint: disable=protected-access

    @override
    def update(self, *args, **kwargs) -> "FromFun":
        """Update function."""
        if self.is_empty:
            return self.create(*args, **kwargs)

        val = self._call_fn(*args, **kwargs)  # pylint: disable=protected-access
        return type(self)(state=self._state.update(*val))

    def compute(self) -> _OutT:
        """Compute function."""
        if self._state is None:
            raise RuntimeError("Nothing to compute, metric is empty!")
        return self._state.compute()

    @classmethod
    def _call_fn(cls, *args, **kwargs) -> tuple:
        """Call fn."""
        val = cls.func(*args, **kwargs)

        # Automatically unroll a tuple of return values
        if not isinstance(val, tuple):
            val = (val,)

        return val



from typing import TYPE_CHECKING

from lightning_utilities.core import rank_zero

if TYPE_CHECKING:
    import reax


class _CheckpointConnector:
    def __init__(self, trainer: "reax.Trainer") -> None:
        """Init function."""
        self._trainer = trainer

    def _get_checkpoint_path(self) -> str | None:
        """Get checkpoint path."""
        checkpointers = self._trainer.checkpoint_listeners
        if not checkpointers:
            return None

        if len(checkpointers) > 1:
            fn = self._get_checkpoint_path
            rank_zero.rank_zero_warn(
                f'`.{fn}(ckpt_path="best")` found Trainer with multiple `ModelCheckpoint`'
                " callbacks. The best checkpoint path from first checkpoint callback will be used."
            )

        checkpointer = checkpointers[0]
        return getattr(checkpointer, "best_model_path", None)



import contextlib
import functools
import logging
import os
from typing import TYPE_CHECKING, Any, TypeVar
import weakref

import beartype
from flax import nnx
import fsspec.utils
import jax
import jaxtyping as jt

from . import data as data_
from . import hooks
from . import loggers as loggers_
from . import modules, optimizers, strategies
from .data import DeviceDataLoader
from .training import _logger_connector
from .utils import events

if TYPE_CHECKING:
    import reax

__all__ = ("Engine",)

_OutT = TypeVar("_OutT")
_LOGGER = logging.getLogger(__name__)


class Engine:
    """
    The central execution and orchestration component for distributed training.

    The Engine serves as the high-level interface that binds a specific
    Strategy (e.g., local, DDP, sharded) to the execution flow,
    providing standardized methods necessary to build training loops or a full
    Trainer.

    It encapsulates the environment setup defined by the chosen strategy,
    managing details like accelerator configuration, process launching (e.g.,
    via MPI or Python's multiprocessing), and backend initialization.

    Attributes:
        strategy (object): The execution strategy defining the environment and distributed
            communication.
    """

    def __init__(
        self,
        accelerator: str = "auto",
        strategy: "str | reax.Strategy" = "auto",
        devices: list[int] | str | int = "auto",
        precision=None,
        logger: "reax.Logger | Iterable[reax.Logger] | bool | None" = True,
        listeners: "list[reax.TrainerListener] | reax.TrainerListener | None" = None,
        deterministic: bool = False,
        rngs: nnx.Rngs = None,
        default_root_dir: "reax.types.Path | None" = None,
    ):
        """
        Initializes the Engine with execution parameters and components.

        The Engine sets up the chosen execution strategy and initializes core
        components like loggers, event listeners, and the random number generator
        state.

        Args:
            accelerator (str): The type of hardware accelerator to use, e.g., ``'gpu'``, ``'tpu'``,
                or ``'cpu'``. Defaults to ``'auto'``.
            strategy (str | reax.Strategy): The distributed training strategy to use, e.g.,
                ``'ddp'``, ``'fsdp'``, or a :class:`reax.Strategy` instance. Defaults to ``'auto'``.
            devices (list[int] | str | int): The specific devices to target. Can be a list of
                indices, ``'auto'``, or an integer count.
            precision: Currently not supported in JAX/nnx, and a warning is issued if set.
            logger (reax.Logger | Iterable[reax.Logger] | bool | None): One or more loggers, or
                ``True`` (for default CSVLogger) or ``False`` (to disable logging).
            listeners (list[reax.TrainerListener] | reax.TrainerListener | None): A single or list
                of :class:`reax.TrainerListener` objects for hooking into events.
            deterministic (bool): Currently not supported, a warning is issued if set.
            rngs (nnx.Rngs): The initial random number generator state used for all training steps.
                Defaults to a new :class:`nnx.Rngs(0)`.
            default_root_dir (typing.Path | None): The default directory path for saving logs and
                checkpoints if not specified elsewhere. Defaults to the current working directory.
        """
        if deterministic:
            _LOGGER.warning("`deterministic=True` is not supported yet, ignoring.")
        if precision is not None:
            _LOGGER.warning(
                "`precision` other than None is not supported yet, ignoring.  "
                "There is still ongoing discussion on how to support this in JAX, "
                "see e.g.: https://github.com/jax-ml/jax/issues/22688"
            )
        # Params

        # State
        if isinstance(strategy, strategies.Strategy):
            self._strategy = strategy
        else:
            self._strategy = strategies.create(strategy, accelerator, devices=devices)
        self._default_root_dir = (
            os.fspath(default_root_dir) if default_root_dir is not None else os.getcwd()
        )
        self._events = events.EventGenerator[hooks.TrainerListener](
            default_args=(weakref.proxy(self),)
        )
        self._rngs = rngs if rngs is not None else nnx.Rngs(0)
        self._loggers: list[loggers_.Logger] = _init_loggers(logger, self.default_root_dir)
        self._logging = _logger_connector.TrainerLogging()

        if isinstance(listeners, hooks.TrainerListener):
            listeners = [listeners]
        if listeners:
            for listener in listeners:
                self._events.add_listener(listener)

    def finalize(self):
        """Clean up the trainer.

        After this is called, this object should no longer be interacted with,
        as internal components (like event listeners and loggers) are cleared.
        """
        self._events = None
        self._loggers = None

    @property
    def default_root_dir(self) -> str:
        """
        Get the fallback directory used for loggers and other components when not explicitly
        specified, resolving path variables (e.g., ``~``) if necessary.

        Returns:
            str: The resolved default root directory path.
        """
        if _is_local_file_protocol(self._default_root_dir):
            return os.path.normpath(os.path.expanduser(self._default_root_dir))

        return self._default_root_dir

    @property
    def logger(self) -> "reax.Logger | None":
        r"""
        Get the first (and main) logger configured for the Engine.

        Returns:
            reax.Logger | None: The main logger, or None if no loggers are configured.
        """
        if not self._loggers:
            return None

        return self._loggers[0]

    @property
    def loggers(self) -> list["reax.Logger"]:
        """
        Get all the loggers configured for the Engine.

        Returns:
            list[reax.Logger]: A list of all configured loggers.
        """
        return self._loggers

    @loggers.setter
    def loggers(self, loggers: list["reax.Logger"] | None) -> None:
        """
        Set the list of loggers.

        Args:
            loggers (list[reax.Logger] | None): The new list of loggers.
        """
        self._loggers = loggers if loggers else []

    @property
    def rngs(self) -> nnx.Rngs:
        """
        The current random number generator state used by the Engine.

        Returns:
            nnx.Rngs: The RNG state object.
        """
        return self._rngs

    @property
    def strategy(self) -> strategies.Strategy:
        """
        The execution strategy currently employed by the Engine.

        Returns:
            strategies.Strategy: The underlying strategy object.
        """
        return self._strategy

    @property
    def is_global_zero(self) -> bool:
        """
        Whether this process is the global rank zero process.

        Returns:
            bool: True if the current process has a global rank of zero.
        """
        return self._strategy.is_global_zero

    @property
    def local_process_index(self) -> int:
        """
        Rank of the process on the current host (0 → local device count − 1).

        Returns:
            int: The local rank.
        """
        return getattr(self.strategy, "local_rank", 0)

    @property
    def node_rank(self) -> int:
        """
        The rank of the current node (host) among all nodes.

        Returns:
            int: The node rank.
        """
        return getattr(self.strategy, "node_rank", 0)

    @property
    def process_index(self) -> int:
        """
        Global rank of the current process across all hosts (0 → total process count − 1).

        Returns:
            int: The global rank.
        """
        return getattr(self.strategy, "process_index", 0)

    @property
    def process_count(self) -> int:
        """
        Total number of processes across all hosts participating in the distributed task.

        Returns:
            int: The total number of processes.
        """
        return getattr(self.strategy, "process_count", 1)

    @property
    def device(self) -> jax.Device:
        """
        The JAX device instance currently assigned to this process.

        Returns:
            jax.Device: The assigned JAX device.
        """
        return self._strategy.device

    @contextlib.contextmanager
    def default_device(self):
        """
        Context manager that explicitly sets the strategy's device as the default for the
        duration of the context, ensuring subsequent operations target this device.

        Yields:
            None: The context manager yields nothing.
        """
        with jax.default_device(self.device):
            yield

    def setup(self, *args) -> Any:
        """
        Prepares and transforms objects (Modules, Optimizers, DataLoaders) for distributed use
        according to the current strategy.

        For example, this method wraps Optimizers in a
        :class:`~reax.optimizers.DistributedOptimizer` and moves Module parameters to the
        correct device.

        Args:
            *args: One or more objects (:class:`~reax.modules.Module`,
                :class:`~reax.optimizers.Optimizer`, or :class:`~reax.data.DataLoader`) to set up.

        Returns:
            Any: The set up object(s), with distributed wrappers applied. If a single object was
            passed, a single object is returned; otherwise, a tuple is returned.
        """
        setup_fn = functools.partial(_setup, engine=self)
        res = tuple(map(setup_fn, args))
        if len(args) == 1:
            return res[0]

        return res

    def setup_dataloaders(
        self, *args
    ) -> "reax.data.DeviceDataLoader | list[reax.data.DeviceDataLoader]":
        """
        Prepares and wraps dataloaders for the distributed environment.

        This delegates to the strategy to potentially shard the data and wraps
        the result in a :class:`~reax.data.DeviceDataLoader` to ensure data is
        moved to the correct device automatically.

        Args:
            *args: One or more :class:`reax.DataLoader` instances.

        Returns:
            reax.data.DeviceDataLoader | list[reax.data.DeviceDataLoader]: The prepared
            dataloader(s), wrapped as a :class:`~reax.data.DeviceDataLoader`.
        """
        loaders = list(map(self.strategy.setup_dataloader, args))
        # Wrap in DeviceLoader so all data is already on the selected device
        loaders = list(map(self._wrap_loader, loaders))
        if len(args) == 1:
            return loaders[0]

        return loaders

    def to_device(self, data: jt.PyTree) -> jt.PyTree:
        """
        Moves the provided PyTree data structure to the Engine's assigned device.

        Args:
            data (jaxtyping.PyTree): The PyTree object (e.g., tensors, arrays) to move.

        Returns:
            jaxtyping.PyTree: The data moved to the device.
        """
        return self._strategy.to_device(data)

    def barrier(self, name: str | None = None) -> None:
        """
        Wait for all processes to enter this call.

        Use this to synchronize all parallel processes, but only if necessary, otherwise the
        overhead of synchronization will cause your program to slow down.
        **This method needs to be called on all processes.** Failing to do so will cause your
        program to stall forever.

        Args:
            name (str | None): An optional name for the barrier for debugging purposes.
        """
        self._strategy.barrier(name=name)

    def broadcast(self, obj: jt.PyTree, src: int = 0) -> jt.PyTree:
        r"""
        Send a tensor from one process to all others.

        **This method needs to be called on all processes.** Failing to do so will cause your
        program to stall forever.

        Args:
            obj (jaxtyping.PyTree): The object to broadcast to all other members.
                Any pytree object is supported.
            src (int): The (global) rank of the process that should send the data to all others.
                Defaults to 0.

        Returns:
            jaxtyping.PyTree: The transferred data, which is the same value on every rank.
        """
        return self._strategy.broadcast(obj, src=src)

    def all_reduce(self, obj: jt.PyTree, reduce_op: str = "mean") -> jt.PyTree:
        """
        Reduces a tensor from several distributed processes to one aggregated tensor (AllReduce).

        The result is the same on all processes.

        Args:
            obj (jaxtyping.PyTree): The pytree to synchronize and reduce.
            reduce_op (str): The reduction operation. Defaults to ``'mean'``/``'avg'``.
            Can also be ``'sum'``.

        Returns:
            jaxtyping.PyTree: The reduced value.
        """
        return self._strategy.all_reduce(obj, reduce_op=reduce_op)

    def call(self, name: str, *args, **kwargs):
        """
        Triggers a named event hook for all registered listeners.

        The event hook is expected to be a method on :class:`reax.hooks.TrainerListener`.

        Args:
            name (str): The name of the event hook method to call (e.g., ``'on_train_epoch_end'``).
            *args: Positional arguments to pass to the hook.
            **kwargs: Keyword arguments to pass to the hook.
        """
        self._events.fire_event(getattr(hooks.TrainerListener, name), *args, **kwargs)

    def compute(self, metric: "reax.typing.MetricInstance[_OutT]") -> _OutT:
        """
        Computes the final value of a metric across all distributed processes.

        Unlike calling ``metric.compute()`` directly, this method ensures the
        value is aggregated correctly across all ranks (e.g., summing up counts)
        before the final calculation.

        Args:
            metric (reax.types.MetricInstance): The metric object to compute the final value for.

        Returns:
            _OutT: The computed metric value.
        """
        return self._strategy.compute(metric)

    def _wrap_loader(self, loader: "reax.DataLoader") -> "reax.data.DeviceDataLoader":
        """
        Internal utility to wrap a standard DataLoader into a DeviceDataLoader.
        """
        if isinstance(loader, DeviceDataLoader):
            loader = loader.parent

        return data_.DeviceDataLoader(loader, self.device)


@functools.singledispatch
def _setup(arg, engine: Engine) -> Any:
    raise TypeError(f"Unknown argument type: {arg.__class__.__name__}")


@_setup.register
def _(arg: optimizers.Optimizer, engine: Engine) -> optimizers.Optimizer:
    return optimizers.DistributedOptimizer(
        opt=arg.optimizer, state=arg.state, engine=engine, count=arg.update_count
    )


@_setup.register
def _(arg: modules.Module, engine: Engine) -> modules.Module:
    return arg.set_parameters(engine.to_device(arg.parameters()))


@_setup.register
def _(arg: data_.DataLoader, engine: Engine) -> data_.DataLoader:
    return engine.strategy.setup_dataloader(arg)


@jt.jaxtyped(typechecker=beartype.beartype)
def _init_loggers(
    logger: "reax.Logger | Iterable[reax.Logger] | bool | None",
    default_root_dir: str,
) -> list["reax.Logger"]:
    """Init loggers."""
    if isinstance(logger, loggers_.Logger):
        return [logger]

    if isinstance(logger, (bool, type(None))):
        if logger:
            return [loggers_.CsvLogger(default_root_dir)]

        return []

    return list(logger)


def _is_local_file_protocol(path: "reax.types.Path") -> bool:
    """Is local file protocol."""
    return fsspec.utils.get_protocol(str(path)) == "file"



import dataclasses
from typing import TYPE_CHECKING, Generic, TypeVar

import jax
import jax.numpy as jnp
import jaxtyping as jt
from typing_extensions import override

from . import metrics as _metric
from . import types

if TYPE_CHECKING:
    import reax


_OutT = TypeVar("_OutT")


@dataclasses.dataclass
class Metadata:
    fx: str
    name: str
    batch_idx: int
    prog_bar: bool
    logger: bool
    on_step: bool
    on_epoch: bool


class ArrayResultMetric(_metric.Metric[jax.Array]):
    value: jax.Array = 0  # Redefine here so the typing hinting works
    cumulated_batch_size: int = 0

    @classmethod
    def create(
        # pylint: disable=arguments-differ
        cls,
        value: jt.ArrayLike,
        batch_size: int,
    ) -> "ArrayResultMetric":
        """Create function."""
        return ArrayResultMetric(value=jnp.asarray(value), cumulated_batch_size=batch_size)

    @override
    def update(
        # pylint: disable=arguments-differ
        self,
        value: jt.ArrayLike,
        batch_size: int,
    ) -> "ArrayResultMetric":
        """Update function."""
        return ArrayResultMetric(
            value=self.value + jnp.asarray(value),
            cumulated_batch_size=self.cumulated_batch_size + batch_size,
        )

    @override
    def merge(self, other: "ArrayResultMetric") -> "ArrayResultMetric":
        """Merge function."""
        return ArrayResultMetric(
            value=self.value + other.value,
            cumulated_batch_size=self.cumulated_batch_size + other.cumulated_batch_size,
        )

    @override
    def compute(self) -> jax.Array:
        """Compute function."""
        return self.value / self.cumulated_batch_size


class ResultEntry(Generic[_OutT]):
    def __init__(
        self,
        meta: Metadata,
        metric: "reax.typing.MetricInstance[_OutT]",
        last_value: "reax.types.MetricInstance[_OutT] | None" = None,
    ):
        """Init function."""
        self._meta = meta  # Readonly
        self.metric = metric
        self._last_value: "reax.types.MetricInstance[_OutT] | _OutT | None" = last_value

    @property
    def meta(self) -> Metadata:
        """Meta function."""
        return self._meta

    @property
    def last_value(self) -> _OutT | None:
        """Last value."""
        if isinstance(self._last_value, types.MetricInstance):
            # Lazily compute the metric as it has now been requested
            self._last_value = self._last_value.compute()

        return self._last_value


class ResultCollection(dict[str, ResultEntry]):
    """A dictionary holding model metrics."""

    def __str__(self) -> str:
        """Str function."""
        my_str = str(self)
        return f"{type(self)}.__name__({my_str})"

    def log(
        self,
        fx: str,
        name: str,
        value: "jt.ArrayLike | reax.typing.MetricInstance[_OutT]",
        batch_idx: int,
        *,
        prog_bar: bool = False,
        logger: bool = False,
        on_step: bool = False,
        on_epoch: bool = True,
        batch_size: int | None = None,
    ):
        """Log function."""
        key = f"{fx}.{name}"

        if isinstance(value, types.MetricInstance):
            metric = value
        else:
            try:
                metric = ArrayResultMetric.create(value, batch_size)
            except TypeError:
                raise TypeError(
                    f"Value must be a `reax.Metric` or a raw value, got {type(value).__name__}"
                ) from None

        meta = Metadata(
            fx=fx,
            name=name,
            batch_idx=batch_idx,
            prog_bar=prog_bar,
            logger=logger,
            on_step=on_step,
            on_epoch=on_epoch,
        )
        last_value = metric

        if key in self:
            # Merge with existing metric to propagate results
            metric = self[key].metric.merge(metric)

        self[key] = ResultEntry(meta, metric, last_value=last_value)



import types

import jax
import jax.numpy as jnp
import jax.typing
import numpy as np


def to_base(array: jax.typing.ArrayLike) -> int | float | list:
    """To base."""
    if isinstance(array, (int, float)):
        return array

    if isinstance(array, list):
        return array

    if jnp.isscalar(array):
        return array.item()

    return array.tolist()


def to_scalar(array: jax.typing.ArrayLike) -> float | int:
    """Convert an array to a python scalar type."""
    if isinstance(array, (np.ndarray, jax.Array)):
        return array.item()

    return array


def infer_backend(pytree) -> types.ModuleType:
    """Try to infer a backend from the passed pytree (numpy or jax.numpy)"""
    any_numpy = any(isinstance(x, np.ndarray) for x in jax.tree_util.tree_leaves(pytree))
    any_jax = any(isinstance(x, jax.Array) for x in jax.tree_util.tree_leaves(pytree))
    if any_numpy and any_jax:
        raise ValueError("Cannot mix numpy and jax arrays")

    if any_numpy:
        return np

    if any_jax:
        return jnp

    return jnp



from . import progress_bar, rich_progress, tqdm_progress
from .progress_bar import *
from .rich_progress import *
from .tqdm_progress import *

__all__ = progress_bar.__all__ + tqdm_progress.__all__ + rich_progress.__all__ + ("rich_progress",)



import functools
from typing import TYPE_CHECKING

from . import _metric, _registry

if TYPE_CHECKING:
    import reax

__all__ = ("get",)


@functools.singledispatch
def get(metric) -> "reax.Metric":
    """Get function."""
    try:
        if issubclass(metric, _metric.Metric):
            return metric()
    except TypeError:
        pass

    raise TypeError(f"Cannot get metric, got {type(metric).__name__}")


@get.register
def get_str(metric: str):
    """Get str."""
    return _registry.get(metric)


@get.register(_metric.Metric)
def get_metric(metric: "reax.Metric"):
    """Get metric."""
    return metric



from typing import TYPE_CHECKING

from . import _jax, _single_device, _utils

if TYPE_CHECKING:
    import reax

__all__ = ("create",)


def create(name: str, platform: str | None, **kwargs) -> "reax.Strategy":
    if name == "auto":
        if "devices" not in kwargs or kwargs["devices"] == "auto":
            num_devices = _utils.probe_local_device_count(platform)
        else:
            num_devices = kwargs["devices"]

        if num_devices > 1:
            name = "ddp"
        else:
            name = "single"

    if name == "single":
        return _single_device.SingleDevice(platform)

    if name == "ddp":
        return _jax.JaxDdpStrategy(platform, **kwargs)

    raise ValueError(f"Unknown strategy: {name}")



from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import reax


__all__ = ("TrainerListener",)


class TrainerListener:
    # pylint: disable=too-many-public-methods

    def on_fit_start(self, trainer: "reax.Trainer", stage: "reax.stages.Fit", /) -> None:
        """A fitting stage is about to begin."""

    def on_fit_end(self, trainer: "reax.Trainer", stage: "reax.stages.Fit", /) -> None:
        """A fitting stage is about to end."""

    def on_train_start(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /):
        """Training is starting"""

    def on_train_epoch_start(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        """A training epoch is about to begin."""

    def on_train_batch_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Train", batch: Any, batch_idx: int, /
    ) -> None:
        """The training stage if about to process a batch."""

    def on_train_batch_end(
        self,
        trainer: "reax.Trainer",
        stage: "reax.stages.Train",
        outputs: Any,
        batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """The training stage has just finished processing a batch."""

    def on_train_epoch_end(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        """A training epoch is about to end."""

    def on_train_end(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /):
        """Training is ending"""

    def on_validation_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Validate", /
    ) -> None:
        """Validation is starting."""

    def on_validation_epoch_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Validate", /
    ) -> None:
        """A validation epoch is about to begin."""

    def on_validation_batch_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Validate", batch: Any, batch_idx: int, /
    ) -> None:
        """The validation stage if about to process a batch."""

    def on_validation_batch_end(
        self,
        trainer: "reax.Trainer",
        stage: "reax.stages.Train",
        outputs: Any,
        batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """The validation stage has just finished processing a batch."""

    def on_validation_epoch_end(
        self, trainer: "reax.Trainer", stage: "reax.stages.Validate", /
    ) -> None:
        """A validation epoch is about to end."""

    def on_validation_end(self, trainer: "reax.Trainer", stage: "reax.stages.Validate", /) -> None:
        """Validation has ended."""

    def on_test_start(self, trainer: "reax.Trainer", stage: "reax.stages.Test", /) -> None:
        """Test is starting."""

    def on_test_epoch_start(self, trainer: "reax.Trainer", stage: "reax.stages.Test", /) -> None:
        """A test epoch is about to begin."""

    def on_test_batch_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Test", batch: Any, batch_idx: int, /
    ) -> None:
        """The test stage if about to process a batch."""

    def on_test_batch_end(
        self,
        trainer: "reax.Trainer",
        stage: "reax.stages.Test",
        outputs: Any,
        batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """The test stage has just finished processing a batch."""

    def on_test_epoch_end(self, trainer: "reax.Trainer", stage: "reax.stages.Test", /) -> None:
        """A test epoch is about to end."""

    def on_test_end(self, trainer: "reax.Trainer", stage: "reax.stages.Test", /) -> None:
        """Test is ending."""

    def on_predict_start(self, trainer: "reax.Trainer", stage: "reax.stages.Predict", /) -> None:
        """Predict is starting."""

    def on_predict_epoch_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Predict", /
    ) -> None:
        """A predict epoch is about to begin."""

    def on_predict_batch_start(
        self, trainer: "reax.Trainer", stage: "reax.stages.Predict", batch: Any, batch_idx: int, /
    ) -> None:
        """The test stage if about to process a batch."""

    def on_predict_batch_end(
        self,
        trainer: "reax.Trainer",
        stage: "reax.stages.Train",
        outputs: Any,
        batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """The predict stage has just finished processing a batch."""

    def on_predict_epoch_end(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        """A predict epoch is about to end."""

    def on_predict_end(self, trainer: "reax.Trainer", stage: "reax.stages.Predict", /) -> None:
        """Predict is ending."""

    def on_epoch_start(self, trainer: "reax.Trainer", stage: "reax.stages.EpochStage", /) -> None:
        """An epoch is just about to begin."""

    def on_batch_end(
        self,
        trainer: "reax.Trainer",
        stage: "reax.stages.EpochStage",
        batch_idx: int,
        metrics: dict,
        /,
    ) -> None:
        """A batch has just been processed."""

    def on_epoch_end(
        self, trainer: "reax.Trainer", stage: "reax.stages.EpochStage", metrics: dict, /
    ) -> None:
        """An epoch is ending."""

    def on_exception(
        self,
        trainer: "reax.Trainer",
        stage: "reax.Stage",
        exception: BaseException,
        /,
    ) -> None:
        """Called when any trainer execution is interrupted by an exception."""

    # region Generic messages

    def setup(self, trainer: "reax.Trainer", stage: "reax.Stage", /) -> None:
        """Called when a stage (e.g. fit, validate, test, predict) begins."""

    def on_stage_start(self, trainer: "reax.Trainer", stage: "reax.stages.Stage", /) -> None:
        """A trainer stage is starting."""

    def on_stage_iter_start(
        self, trainer: "reax.Trainer", stage: "reax.Stage", step: int, /
    ) -> None:
        """A stage is about to start an interation."""

    def on_stage_iter_end(
        self, trainer: "reax.Trainer", stage: "reax.Stage", step: int, outputs: Any, /
    ):
        """The stage just finished processing an iteration."""

    def on_stage_end(self, trainer: "reax.Trainer", stage: "reax.Stage", /) -> None:
        """The stage is ending."""

    def teardown(self, trainer: "reax.Trainer", state: "reax.Stage", /) -> None:
        """Called when a stage (e.g. fit, validate, test, predict, or tune ends)."""

    # endregion



from typing import TYPE_CHECKING, Any, TypeVar
import weakref

import beartype
from flax import nnx
import jaxtyping as jt
import optax
from typing_extensions import override

from reax import exceptions
from reax import optimizers as optimizers_
from reax.lightning import rank_zero

from . import common, stages

if TYPE_CHECKING:
    import reax


__all__ = ("Train",)


_T_co = TypeVar("_T_co", covariant=True)


class Train(stages.EpochStage):
    """One training epoch."""

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        module: "reax.Module",
        datamanager: "reax.data.DataSourceManager",
        engine: "reax.Engine",
        optimizers: "list[reax.Optimizer]",
        *,
        rngs: nnx.Rngs | None = None,
        fast_dev_run: bool | int = False,
        min_updates: int = 0,
        max_updates: int | float | None = None,
        limit_batches: int | float | None = None,
        accumulate_grad_batches: int = 1,
        stopper: common.Stopper | None = None,
    ):
        super().__init__(
            "fit",
            module,
            datamanager,
            engine,
            rngs=rngs,
            dataloader_name="train",
            fast_dev_run=fast_dev_run,
            limit_batches=limit_batches,
            enable_checkpointing=True,
        )
        # Params
        self._min_updates = min_updates
        self._max_updates = max_updates
        self._accumulate_grad_batches = accumulate_grad_batches

        # State
        self._optimizers = optimizers
        self._stopper = stopper if stopper is not None else common.Stopper()
        self._stopper.add_condition(lambda: self.updates >= self._min_updates)

    @property
    def num_training_batches(self) -> int | float | None:
        return self.max_batches

    @property
    def updates(self) -> int:
        """Get the number of gradient updates that have been applied."""
        return sum(opt.update_count for opt in self._optimizers)

    @property
    def optimizers(self) -> list["reax.Optimizer"] | None:
        """Optimizers function."""
        return self._optimizers

    @override
    def run(self) -> list["reax.Optimizer"]:
        """Run function."""
        super().run()
        return self._optimizers

    @override
    def _on_starting(self):
        """On starting."""
        super()._on_starting()

        if not self._optimizers:
            opts = self._module.configure_optimizers()
            if opts is None:
                rank_zero.rank_zero_warn(
                    "`reax.Module.configure_optimizers` returned `None`, this fit will run "
                    "with no optimizer"
                )
                opt = optimizers_.mock_optimizer
                opts = opt, opt.init(self._module.parameters())

            if not isinstance(opts, list):
                opts = [opts]

            optimizers: list[optimizers_.Optimizer] = []
            for opt, state in opts:
                # Move optimizer parameters to device
                state = self._engine.to_device(state)
                if self._accumulate_grad_batches > 1:
                    stepper = optax.MultiSteps(opt, every_k_schedule=self._accumulate_grad_batches)
                    state = stepper.init(self._module.parameters())
                    opt = stepper.gradient_transformation()

                optimizers.append(optimizers_.Optimizer(opt, state))

            # Create the `Optimizer` instances
            self._optimizers = optimizers

        self._module.on_train_start(weakref.proxy(self))

    @override
    def _on_iteration_starting(self):
        super()._on_iteration_starting()
        self._module.on_train_batch_start(weakref.proxy(self), self.batch, self.batch_idx)

    @override
    def _on_epoch_start(self):
        """On started."""
        super()._on_epoch_start()
        self._module.on_train_epoch_start(weakref.proxy(self))

    @override
    def _step(self) -> Any:
        """Step function."""
        if self._module.parameters() is None:
            raise exceptions.MisconfigurationException(
                "Module does not have any parameters set, this should have been done in "
                ".configure_model()."
            )

        res = self._module.training_step(self.batch, self._iter)
        if self._module.automatic_optimization:
            if isinstance(res, dict):
                grad = res["grad"]
                loss = None
            else:
                loss, grad = res
            opt = self._optimizers[0]
            self._module.on_before_optimizer_step(opt, grad)
            opt = opt.update_module(self._module, grad, value=loss)
            self._optimizers = [opt]

        if (self._min_updates is None or self.updates >= self._min_updates) and (
            self._max_updates is not None and self.updates >= self._max_updates
        ):
            self.stop("Max updates reached")

        return res

    @override
    def _on_iteration_finishing(self, outputs: Any, /):
        super()._on_iteration_finishing(outputs)
        self._module.on_train_batch_end(weakref.proxy(self), outputs, self.batch, self.batch_idx)

    @override
    def _on_epoch_end(self) -> None:
        super()._on_epoch_end()
        self._module.on_train_epoch_end(weakref.proxy(self))

    @override
    def _on_stopping(self) -> None:
        super()._on_stopping()
        self._module.on_train_end(weakref.proxy(self))

    @override
    def _done(self) -> bool:
        """Done function."""
        if self.max_batches is not None and self.batch_idx >= self.max_batches:
            rank_zero.rank_zero_debug(
                f"`{type(self).__name__}` done: max_batches.{self.max_batches!r}` reached."
            )
            return True

        if self._max_updates is not None and self.updates >= self._max_updates:
            rank_zero.rank_zero_debug(
                f"`{type(self).__name__}` done: `max_updates={self._max_updates!r}` reached."
            )
            return True

        if self._stopper.stop_requested:
            if self._stopper.can_stop:
                rank_zero.rank_zero_debug(
                    f"`{type(self).__name__}` stopped: `{type(self).__name__}.should_stop` was set."
                )
            else:
                self._warning_cache.info(
                    f"Trainer was signaled to stop but the required "
                    f"`min_epochs={self.parent.min_iters!r}` or"
                    f" `min_steps={self._min_updates!r}` has not been met. "
                    f"Training will continue..."
                )
            return self._stopper.can_stop

        return False



"""Test loop."""

import logging
from typing import TYPE_CHECKING
import weakref

import beartype
from flax import nnx
import jaxtyping as jt
from typing_extensions import override

from . import stages

if TYPE_CHECKING:
    import reax

__all__ = ("Test",)

_LOGGER = logging.getLogger(__name__)


class Test(stages.EpochStage):
    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        module: "reax.Module",
        datamanager: "reax.data.DataSourceManager",
        engine: "reax.Engine",
        *,
        rngs: nnx.Rngs = None,
        fast_dev_run: bool | int = False,
        limit_batches: int | float | None = None,
    ):
        """Init function."""
        super().__init__(
            "test",
            module,
            datamanager,
            engine,
            rngs=rngs,
            fast_dev_run=fast_dev_run,
            limit_batches=limit_batches,
        )

    @override
    def _on_starting(self):
        super()._on_starting()
        self._module.on_test_start(weakref.proxy(self))

    @override
    def _on_epoch_start(self):
        super()._on_epoch_start()
        self._module.on_test_epoch_start(weakref.proxy(self))

    @override
    def _step(self) -> "reax.stages.MetricResults":
        return self._module.test_step(self.batch, self._iter)

    @override
    def _on_epoch_end(self):
        super()._on_epoch_end()
        self._module.on_test_epoch_end(weakref.proxy(self))

    @override
    def _on_stopping(self) -> None:
        super()._on_stopping()
        self._module.on_test_end(weakref.proxy(self))



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/9177ec09caadcf88859e1f1e3e10a18e8832069a/src/lightning/fabric/loggers/csv_logs.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CSV logger
----------

CSV logger for basic experiment logging that does not require opening ports

"""

from argparse import Namespace
from collections.abc import Mapping
import csv
import logging
import os
from typing import Any

import fsspec
import jax
from typing_extensions import override

from reax.lightning import rank_zero

from . import _utils, logger
from .. import saving, types

_LOGGER = logging.getLogger(__name__)

__all__ = "CsvLogger", "CSVLogger"


class CsvLogger(logger.Logger):
    r"""Log to local file system in yaml and CSV format.

    Logs are saved to ``os.path.join(save_dir, name, version)``.

    Example:
        >>> import reax
        >>> logger = reax.loggers.CsvLogger("logs", name="my_exp_name")
        >>> trainer = reax.Trainer(logger=logger)

    Args:
        save_dir: Save directory
        name: Experiment name, optional. Defaults to ``'reax_logs'``. If name is ``None``, logs
            (versions) will be stored to the save dir directly.
        version: Experiment version. If version is not specified the logger inspects the save
            directory for existing versions, then automatically assigns the next available version.
        prefix: A string to put at the beginning of metric keys.
        flush_logs_every_n_steps: How often to flush logs to disk (defaults to every 100 steps).

    """

    LOGGER_JOIN_CHAR = "-"

    def __init__(
        self,
        save_dir: types.Path,
        name: str | None = "reax_logs",
        version: int | str | None = None,
        prefix: str = "",
        flush_logs_every_n_steps: int = 100,
    ):
        super().__init__()
        save_dir = os.fspath(save_dir)
        self._root_dir = save_dir
        self._name = name or ""
        self._version = version
        self._prefix = prefix
        self._fs: fsspec.AbstractFileSystem = fsspec.url_to_fs(save_dir)[0]
        self._experiment: ExperimentWriter | None = None
        self._flush_logs_every_n_steps = flush_logs_every_n_steps

    @property
    @override
    def name(self) -> str:
        """Gets the name of the experiment.

        Returns:
            The name of the experiment.

        """
        return self._name

    @property
    @override
    def version(self) -> int | str:
        """Gets the version of the experiment.

        Returns:
            The version of the experiment if it is specified, else the next version.

        """
        if self._version is None:
            self._version = self._get_next_version()
        return self._version

    @property
    @override
    def root_dir(self) -> str:
        """Gets the save directory where the versioned CSV experiments are saved."""
        return self._root_dir

    @property
    @override
    def log_dir(self) -> str:
        """The log directory for this run.

        By default, it is named ``'version_${self.version}'`` but it can be overridden by passing a
        string value for the constructor's version parameter instead of ``None`` or an int.

        """
        # create a pseudo standard path
        version = self.version if isinstance(self.version, str) else f"version_{self.version}"
        return os.path.join(self._root_dir, self.name, version)

    @property
    @override
    def save_dir(self) -> str:
        """The current directory where logs are saved.

        Returns:
            The path to current directory where logs are saved.

        """
        return self._root_dir

    @property
    @logger.rank_zero_experiment
    def experiment(self) -> "ExperimentWriter":
        """Actual ExperimentWriter object. To use ExperimentWriter features anywhere in your code,
        do the following.

        Example::

            self.logger.experiment.some_experiment_writer_function()

        """
        if self._experiment is not None:
            return self._experiment

        self._fs.makedirs(self.root_dir, exist_ok=True)
        self._experiment = ExperimentWriter(log_dir=self.log_dir)
        return self._experiment

    @override
    @rank_zero.rank_zero_only
    def log_hyperparams(self, params: dict[str, Any] | Namespace, *_, **__) -> None:
        params = _utils.convert_params(params)
        self.experiment.log_hparams(params)

    @override
    @rank_zero.rank_zero_only
    def log_metrics(
        self, metrics: Mapping[str, jax.typing.ArrayLike], step: int | None = None
    ) -> None:
        metrics = _utils.add_prefix(metrics, self._prefix, self.LOGGER_JOIN_CHAR)
        if step is None:
            step = len(self.experiment.metrics)
        self.experiment.log_metrics(metrics, step)
        if (step + 1) % self._flush_logs_every_n_steps == 0:
            self.save()

    @override
    @rank_zero.rank_zero_only
    def save(self) -> None:
        super().save()
        self.experiment.save()

    @override
    @rank_zero.rank_zero_only
    def finalize(self, _status: str) -> None:
        if self._experiment is None:
            # When using multiprocessing, finalize() should be a no-op on the main process, as no
            # experiment has been initialized there
            return
        self.save()

    def _get_next_version(self) -> int:
        versions_root = os.path.join(self._root_dir, self.name)

        if not self._fs.isdir(versions_root):
            return 0

        existing_versions = []
        for d in self._fs.listdir(versions_root):
            full_path = d["name"]
            name = os.path.basename(full_path)
            if self._fs.isdir(full_path) and name.startswith("version_"):
                dir_ver = name.split("_")[1]
                if dir_ver.isdigit():
                    existing_versions.append(int(dir_ver))

        if len(existing_versions) == 0:
            return 0

        return max(existing_versions) + 1


CSVLogger = CsvLogger  # Alias for compatibility with lightning


class ExperimentWriter:
    r"""Experiment writer for CSVLogger.

    Args:
        log_dir: Directory for the experiment logs

    """

    NAME_METRICS_FILE = "metrics.csv"
    NAME_HPARAMS_FILE = "hparams.yaml"

    def __init__(self, log_dir: str) -> None:
        # Params
        self.log_dir = log_dir

        self._fs = fsspec.url_to_fs(log_dir)[0]
        self.metrics_file_path = os.path.join(self.log_dir, self.NAME_METRICS_FILE)

        self._check_log_dir_exists()
        self._fs.makedirs(self.log_dir, exist_ok=True)

        # State
        self.metrics: list[dict[str, float]] = []
        self.metrics_keys: list[str] = []
        self.hparams: dict[str, Any] = {}

    def log_metrics(self, metrics_dict: dict[str, float], step: int | None = None) -> None:
        """Record metrics."""

        def _handle_value(value: jax.Array | Any) -> Any:
            if isinstance(value, jax.Array):
                return value.item()
            return value

        if step is None:
            step = len(self.metrics)

        metrics = {k: _handle_value(v) for k, v in metrics_dict.items()}
        metrics["step"] = step
        self.metrics.append(metrics)

    def log_hparams(self, params: dict[str, Any]) -> None:
        """Record hparams."""
        self.hparams.update(params)

    def save(self) -> None:
        """Save recorded hparams and metrics into files."""
        hparams_file = os.path.join(self.log_dir, self.NAME_HPARAMS_FILE)
        saving.save_hparams_to_yaml(hparams_file, self.hparams)

        if not self.metrics:
            return

        new_keys = self._record_new_keys()
        file_exists = self._fs.isfile(self.metrics_file_path)

        if new_keys and file_exists:
            # we need to re-write the file if the keys (header) change
            self._rewrite_with_new_header(self.metrics_keys)

        with self._fs.open(
            self.metrics_file_path, mode=("a" if file_exists else "w"), newline=""
        ) as file:
            writer = csv.DictWriter(file, fieldnames=self.metrics_keys)
            if not file_exists:
                # only write the header if we're writing a fresh file
                writer.writeheader()
            writer.writerows(self.metrics)

        self.metrics = []  # reset

    def _record_new_keys(self) -> set[str]:
        """Records new keys that have not been logged before."""
        current_keys = set().union(*self.metrics)
        new_keys = current_keys - set(self.metrics_keys)
        self.metrics_keys.extend(new_keys)
        self.metrics_keys.sort()
        return new_keys

    def _rewrite_with_new_header(self, fieldnames: list[str]) -> None:
        with self._fs.open(self.metrics_file_path, "r", newline="") as file:
            metrics = list(csv.DictReader(file))

        with self._fs.open(self.metrics_file_path, "w", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(metrics)

    def _check_log_dir_exists(self) -> None:
        if self._fs.exists(self.log_dir) and self._fs.listdir(self.log_dir):
            rank_zero.rank_zero_warn(
                f"Experiment logs directory {self.log_dir} exists and is not empty."
                " Previous log files in this directory will be deleted when the new ones are saved!"
            )
            if self._fs.isfile(self.metrics_file_path):
                self._fs.rm_file(self.metrics_file_path)



"""Collection of built-in listeners."""

from . import checkpointer, early_stopping, model_checkpoint, progress, utils
from .checkpointer import *
from .early_stopping import *
from .model_checkpoint import *
from .progress import *

__all__ = (
    checkpointer.__all__
    + early_stopping.__all__
    + progress.__all__
    + model_checkpoint.__all__
    + ("utils",)
)



from collections.abc import Callable
from typing import TYPE_CHECKING, ClassVar, Optional, Protocol, TypeVar

import beartype
import clu.internal.utils
import equinox
import jax
import jax.numpy as jnp
import jaxtyping as jt
import numpy as np
from typing_extensions import override

from .. import types
from .. import utils as reax_utils
from ._metric import Metric

if TYPE_CHECKING:
    import reax

__all__ = tuple()

M = TypeVar("M", bound=Metric)
OptionalMask = Optional["reax.types.ArrayMask"]


class ReduceFn(Protocol):
    def __call__(self, values: jax.Array, where: jax.Array = None) -> jax.Array:
        """Perform reduction on the passed values."""


def _prepare_mask(mask: types.ArrayMask, array: jt.Float[jt.Array, "..."]) -> types.ArrayMask:
    """Prepare a mask for use with jnp.where(mask, array, ...).

    This needs to be done to make sure the mask is of the right shape to be compatible with such an
    operation.  The other alternative is

    .. code-block::

        jnp.where(mask, array.T, ...).T

    but this sometimes leads to creating a copy when doing one or both of the transposes.  I'm not
    sure why, but this approach seems to avoid the problem.

    Args:
        mask (jt.Bool[jax.Array, "n_elements"]): The mask to prepare.
        array (jt.Float[jax.Array, "..."]): The array the mask will be
            applied to.

    Returns:
        The prepared mask, typically this is just padded with extra
        dimensions (or reduced).
    """
    return mask.reshape(-1, *(1,) * len(array.shape[1:]))


@jt.jaxtyped(typechecker=beartype.beartype)
def prepare_mask(
    values: jax.Array | np.ndarray,
    mask: OptionalMask = None,
    *,
    return_count: bool = False,
) -> "OptionalMask | tuple[OptionalMask, int | jt.Int[jax.typing.ArrayLike, '']]":
    """Prepare a mask for use with jnp.where(mask, array, ...).  This needs to be done to make sure
    the mask is of the right shape to be compatible with such an operation.  The other alternative
    is
        ``
        jnp.where(mask, array.T, ...).T
        ``
    but this sometimes leads to creating a copy when doing one or both of the transposes.  I'm not
    sure why, but this approach seems to avoid the problem.

    Args:
        values: the array the mask will be applied to
        mask: the mask to prepare
        return_count: is ``True``, returns a tuple where the second
            element is the number of masked elements

    Returns:
        the prepared mask, typically this is just padded with extra
        dimensions (or reduced)
    """
    if mask is None:
        if return_count:
            return None, values.size
        return None

    if mask.shape != values.shape:
        # Create a mask of the correct shape
        if len(mask.shape) > 1:
            raise ValueError(
                "Mask must either have same shape as values array ({values.shape}) or the same "
                "leading dimension, got {mask.shape}."
            )

        mask = _prepare_mask(mask, values)

        # Leading dimensions of mask and predictions must match.
        if mask.shape[0] != values.shape[0]:
            raise ValueError(
                f"Argument `mask` must have the same leading dimension as `values`. "
                f"Received mask of dimension {mask.shape} "
                f"and values of dimension {values.shape}."
            )

    mask = mask.astype(bool)
    clu.internal.utils.check_param(mask, dtype=bool, ndim=values.ndim)
    if return_count:
        np_ = reax_utils.arrays.infer_backend(mask)
        # Calculate the number of non-masked elements in total
        count = values.size if mask is None else np_.array([mask.sum(), *values.shape[1:]]).prod()
        return mask, count

    return mask


def concat(tensors: tuple[jax.Array]) -> jax.Array:
    """Concat function."""
    return jnp.concatenate(tuple(map(jnp.atleast_1d, tensors)))


class WithAccumulator(equinox.Module):
    """Abstract aggregation metric."""

    Self = TypeVar("Self", bound="WithAccumulator")

    reduce_fn: ClassVar[ReduceFn]
    merge_fn: ClassVar[Callable] = concat

    accumulator: jax.Array | None = None

    @classmethod
    def empty(cls):
        return cls(accumulator=None)

    @classmethod
    def create(
        cls,
        values: jax.typing.ArrayLike,
        mask: OptionalMask = None,
    ) -> Self:
        """Create function."""
        mask = prepare_mask(values, mask)
        return cls(accumulator=cls.reduce_fn(values, where=mask))

    @jt.jaxtyped(typechecker=beartype.beartype)
    def merge(self, other: Self) -> Self:
        """Merge function."""
        if self.accumulator is None:
            return other

        cls = type(self)
        return cls(accumulator=cls.reduce_fn(cls.merge_fn((self.accumulator, other.accumulator))))

    @jt.jaxtyped(typechecker=beartype.beartype)
    def update(
        self,
        values: jax.typing.ArrayLike,
        mask: jax.typing.ArrayLike | None = None,
    ) -> Self:
        """Update function."""
        cls = type(self)
        if self.accumulator is None:
            return cls.create(values, mask=mask)

        mask = prepare_mask(values, mask)
        reduced = cls.reduce_fn(values, where=mask)
        return cls(accumulator=cls.reduce_fn(cls.merge_fn((self.accumulator, reduced))))

    @jt.jaxtyped(typechecker=beartype.beartype)
    def compute(self) -> jax.Array:
        """Compute function."""
        return self.accumulator


class WithAccumulatorAndCount(WithAccumulator):
    """Helper class to group common functionality for metrics that can keep track using a total and
    count accumulators.
    """

    Self = TypeVar("Self", bound="WithAccumulator")

    count: jax.Array = 0

    @override
    @classmethod
    def create(
        cls,
        values: jt.ArrayLike,
        mask: OptionalMask = None,
    ) -> Self:
        """Create function."""
        values = jnp.atleast_1d(values)
        mask, num_elements = prepare_mask(values, mask, return_count=True)
        return cls(accumulator=cls.reduce_fn(values, where=mask), count=num_elements)

    @override
    @jt.jaxtyped(typechecker=beartype.beartype)
    def merge(self, other: Self) -> Self:
        """Merge function."""
        if self.count == 0:
            return other

        cls = type(self)
        return cls(
            accumulator=cls.reduce_fn(cls.merge_fn((self.accumulator, other.accumulator))),
            count=self.count + other.count,
        )

    @override
    @jt.jaxtyped(typechecker=beartype.beartype)
    def update(
        self,
        values: jax.typing.ArrayLike,
        mask: jax.typing.ArrayLike | None = None,
    ) -> Self:
        """Update function."""
        cls = type(self)
        if self.count == 0:
            # This metric is empty
            return cls.create(values, mask=mask)

        mask, num_elements = prepare_mask(values, mask, return_count=True)
        reduced = cls.reduce_fn(values, where=mask)
        return cls(
            accumulator=cls.reduce_fn(cls.merge_fn((self.accumulator, reduced))),
            count=self.count + num_elements,
        )

    @override
    @jt.jaxtyped(typechecker=beartype.beartype)
    def compute(self) -> jax.Array:
        """Compute function."""
        return self.accumulator / self.count



import abc
from typing import TYPE_CHECKING

from lightning_utilities.core import overrides

from . import _datasources, datamodules
from .. import modules

if TYPE_CHECKING:
    import reax


___all__ = ("DataSourceManager", "create_manager")


class DataSourceManager(abc.ABC):
    """Manager for coordinating getting data from a source"""

    def __init__(
        self, source: _datasources.DataSource | None, engine: "reax.Engine" = None, **loaders
    ):
        self._datasource: _datasources.DataSource | None = source
        self._engine = engine
        self._loaders: dict[str, "reax.data.DataLoader"] = {
            name: self._setup_dataloader(loaders) for name, loaders in loaders.items()
        }
        self._from_datasource: dict[str, "reax.data.DataLoader"] = {}

    @property
    def _source_base_type(
        self,
    ) -> "type[reax.DataModule] | type[reax.Module]":
        if isinstance(self._datasource, datamodules.DataModule):
            return datamodules.DataModule
        if isinstance(self._datasource, modules.Module):
            return modules.Module

        raise RuntimeError("expected datasource to be a DataModule or Module")

    def has_dataloader(self, name: str) -> bool:
        """Returns `True` if this source provides the name dataloader, `False` otherwise"""
        if name in self._loaders:
            return True

        if name in self._from_datasource:
            return True

        return overrides.is_overridden(
            f"{name}_dataloader", self._datasource, self._source_base_type
        )

    def get_dataloader(self, name: str) -> "reax.data.DataLoader":
        """Get the dataloader of the given name"""
        try:
            return self._loaders[name]
        except KeyError:
            pass

        try:
            return self._from_datasource[name]
        except KeyError:
            pass

        loader = self._request_dataloader(name)
        self._from_datasource[name] = loader
        return loader

    @property
    def source(self) -> "reax.data.DataSource | None":
        """The original source of the dataloader (if there is one)"""
        return self._datasource

    def prepare_data(self) -> None:
        """Tell the data source to prepare the data for use"""
        if self._datasource is not None and overrides.is_overridden(
            "prepare_data", self._datasource, self._source_base_type
        ):
            self._datasource.prepare_data()

    def setup(self, stage) -> None:
        """Tell the data source to set itself up"""
        if self._datasource is not None:
            self._datasource.setup(stage)

    def on_exception(self, exception: BaseException) -> None:
        """Tell the data source that an exception has occurred"""

        if self._datasource is not None:
            self._datasource.on_exception(exception)

    def teardown(self, stage) -> None:
        """Tell the data source to tear down"""

        if self._datasource is not None:
            self._datasource.teardown(stage)

    def reset(self):
        """Reset the cache so dataloaders get reloaded"""
        self._from_datasource = {}

    def _request_dataloader(self, name: str) -> "reax.DataLoader":
        """Get the dataloader directly from the source"""
        loader_name = f"{name}_dataloader"
        loader = getattr(self._datasource, loader_name)()
        loader = self._setup_dataloader(loader)
        return loader

    def _setup_dataloader(self, loader: "reax.DataLoader") -> "reax.DataLoader":
        if self._engine is None:
            return loader

        return self._engine.setup_dataloaders(loader)


def create_manager(
    module: _datasources.DataSource = None,
    datamodule: datamodules.DataModule = None,
    engine: "reax.Engine" = None,
    **loaders,
) -> DataSourceManager:
    # Filter out any loaders that are `None` as this makes calling this method easier
    passed_loaders = {name: loader for name, loader in loaders.items() if loader is not None}

    if datamodule is not None:
        source = datamodule
    elif module is not None:
        source = module
    else:
        source = None

    return DataSourceManager(source, engine, **passed_loaders)



# Copyright (C) 2024  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://raw.githubusercontent.com/Lightning-AI/pytorch-lightning/61a8391617b3f2b5083852ea49b1ead4c2ef72d7/src/lightning/fabric/utilities/rank_zero.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities that can be used for calling functions on a particular rank."""

import logging
import os
from typing import Optional

import lightning_utilities.core.rank_zero as rank_zero_module

# note: we want to keep these indirections so the `rank_zero_only.rank` is set on import
from lightning_utilities.core.rank_zero import (  # noqa: F401
    WarningCache,
    rank_prefixed_message,
    rank_zero_debug,
    rank_zero_deprecation,
    rank_zero_info,
    rank_zero_warn,
)
from typing_extensions import ParamSpec

rank_zero_module.log = logging.getLogger(__name__)


def _get_rank() -> int | None:
    """Get rank."""
    # SLURM_PROCID can be set even if SLURM is not managing the multiprocessing,
    # therefore LOCAL_RANK needs to be checked first
    rank_keys = ("RANK", "LOCAL_RANK", "SLURM_PROCID", "JSM_NAMESPACE_RANK")
    for key in rank_keys:
        rank = os.environ.get(key)
        if rank is not None:
            return int(rank)
    # None to differentiate whether an environment variable was set at all
    return None


rank_zero_only = rank_zero_module.rank_zero_only
# add the attribute to the function but don't overwrite in case Trainer has already set it
rank_zero_only.rank = getattr(rank_zero_only, "rank", _get_rank() or 0)


class ReaxDeprecationWarning(DeprecationWarning):
    """Deprecation warnings raised by Lightning."""


rank_zero_module.rank_zero_deprecation_category = ReaxDeprecationWarning



import logging
import os
import random
import socket
import subprocess  # nosec  # Suppresses Bandit's B404 subprocess warning
import sys
from typing import TYPE_CHECKING, Any, TypeVar

import equinox as eqx
import jax
from jax import distributed, tree
from jax.experimental import multihost_utils
import jax.numpy as jnp
import jaxtyping as jt
from typing_extensions import override

from . import _parallel

if TYPE_CHECKING:
    import reax


__all__ = ("JaxDdpStrategy",)

_LOGGER = logging.getLogger(__name__)


_OutT = TypeVar("_OutT")
Children = list[subprocess.Popen]


class JaxDdpStrategy(_parallel.ParallelStrategy):
    """This strategy uses multi-processing and the JAX library for communication."""

    def __init__(self, platform: str = None, devices: int | str = "auto"):
        _LOGGER.info("Starting JAX DDP strategy...")
        res = self._init(platform, devices)
        self._process_id = res[0]
        self._num_processes: int = res[1]
        self._children = res[2]
        self._device = jax.local_devices()[0]

    @override
    def teardown(self):
        # Wait for all children to finish
        if self._process_id is not None:
            jax.distributed.shutdown()
            for child in self._children:
                child.wait()
            self._children = None
            self._process_id = None

    @staticmethod
    def probe_local_device_count() -> int:
        """Use a subprocess to import JAX and ask it the number of local devices"""
        code = "import jax; print(jax.local_device_count())"
        result = subprocess.check_output([sys.executable, "-c", code])  # nosec # Disable Bandit
        return int(result.decode().strip())

    @staticmethod
    def get_available_port(min_port=49152, max_port=65535, max_attempts=100) -> int:
        """Generates a random port and checks if it's available.  Retries if necessary."""
        for _ in range(max_attempts):
            port = random.randint(min_port, max_port)  # nosec
            try:
                # Attempt to bind to the port.  If it's in use, this will raise an exception.
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.bind(("localhost", port))  # Bind to localhost for testing
                s.close()
                return port
            except OSError:
                # Port is already in use.  Try again.
                pass  # Continue to the next iteration of the loop

        raise OSError(f"Could not find an available port after {max_attempts} attempts.")

    def _init(self, platform: str | None, devices: int | str) -> tuple[int, int, Children]:
        if platform:
            jax.config.update("jax_platforms", platform)

        children = []
        if "REAX_PROCESS_ID" in os.environ:
            env = os.environ
            kwargs = dict(
                coordinator_address=env["REAX_COORDINATOR_ADDRESS"],
                num_processes=int(env["REAX_NUM_PROCESSES"]),
                process_id=int(env["REAX_PROCESS_ID"]),
                local_device_ids=int(env["REAX_PROCESS_ID"]),
            )
            res = self._distributed_init(**kwargs)
            return *res, children

        try:
            # First, let's try the automated method which uses automated cluster detection
            res = self._distributed_init(num_processes=None if devices == "auto" else devices)
        except ValueError:
            # Not in a cluster environment, so open our own processes
            port = self.get_available_port()
            coordinator_address = f"localhost:{port}"
            if devices == "auto":
                num_processes = self.probe_local_device_count()
            else:
                num_processes = devices

            # Launch the other processes
            children = self._launch_children(platform, num_processes, coordinator_address)

            # and now join them
            res = self._distributed_init(
                coordinator_address=coordinator_address,
                num_processes=num_processes,
                process_id=0,
                local_device_ids=0,
            )

        return *res, children

    def _distributed_init(self, **kwargs) -> tuple[int, int]:
        distributed.initialize(**kwargs)
        return jax.process_index(), jax.process_count()

    def _launch_children(
        self, platform: str | None, num: int, coordinator_address: str
    ) -> Children:
        children = []
        for process_id in range(1, num):
            kwargs = {
                "REAX_COORDINATOR_ADDRESS": coordinator_address,
                "REAX_NUM_PROCESSES": str(num),
                "REAX_PROCESS_ID": str(process_id),
            }
            if platform is not None:
                kwargs["JAX_PLATFORMS"] = platform

            # Copy the current environment
            env = os.environ.copy()
            # Add or override variables
            env.update(kwargs)

            # Relaunch the same script with the same Python executable
            if "REAX_COMMAND" in os.environ:
                cmd = os.environ["REAX_COMMAND"].split()
            else:
                cmd = [sys.executable] + sys.argv
            _LOGGER.info("Launching subprocess with: %s", " ".join(cmd))
            children.append(
                subprocess.Popen(  # pylint: disable=consider-using-with
                    cmd, env=env
                )  # nosec # Disable bandit
            )

        return children

    @property
    def device(self):
        return self._device

    @property
    def process_count(self) -> int:
        """Return the total number of ranks."""
        return self._num_processes

    @property
    def process_index(self) -> int:
        return self._process_id

    @override
    def to_device(self, value: Any) -> Any:
        """To device."""
        return jax.device_put(value, self._device)

    @override
    def from_device(self, value: Any) -> Any:
        """From device."""
        return jax.device_get(value)

    @property
    @override
    def is_global_zero(self) -> bool:
        """Is global zero."""
        return self._process_id == 0

    @override
    def broadcast(self, obj: jt.PyTreeDef, src: int = 0) -> Any:
        """Broadcasts an object to all processes.

        Args:
            obj: The pytree to broadcast.
            src (int, optional): Source rank, defaults to 0.
        """
        is_str = isinstance(obj, str)
        if is_str:
            # Encode
            with jax.default_device(self._device):
                obj = jnp.array(list(obj.encode("utf-8")), dtype=jnp.uint8)

        res = multihost_utils.broadcast_one_to_all(obj, src == self.process_index)
        if is_str:
            # Decode
            res = bytes(res.tolist()).decode("utf-8")

        return res

    @override
    def all_gather(self, obj: jt.PyTreeDef) -> Any:
        return multihost_utils.process_allgather(obj)

    @override
    def all_reduce(self, obj: jt.PyTree, reduce_op: str = "mean") -> jt.PyTree:
        """Reduces a tensor from several distributed processes to one aggregated tensor.

        Args:
            obj: the pytree to sync and reduce
            reduce_op: the reduction operation. Defaults to 'mean'/'avg'.
                Can also be a string 'sum' to calculate the sum during reduction.

        Return:
            reduced value
        """
        return getattr(jnp, reduce_op)(self.all_gather(obj))

    @override
    def barrier(self, name: str | None = None) -> None:
        """Synchronizes all processes which blocks processes until the whole group enters this
        function.

        Args:
            name: an optional name to pass into barrier.
        """
        multihost_utils.sync_global_devices(name)

    @override
    def compute(self, metric: "reax.typing.MetricInstance[_OutT]") -> _OutT:
        if self.process_count == 1:
            return metric.compute()

        dynamic, static = eqx.partition(metric, eqx.is_array)
        gathered = multihost_utils.process_allgather(dynamic)
        unbatched: "list[reax.typing.MetricInstance[_OutT]]" = unbatch_pytree(gathered, dynamic)
        # Rejoin with the static data
        unbatched = [eqx.combine(entry, static) for entry in unbatched]

        metric = unbatched[0]
        for entry in unbatched[1:]:
            metric = metric.merge(entry)

        return metric.compute()


def unbatch_pytree(batched: jt.PyTree, original: jt.PyTree) -> list[jt.PyTree]:
    """Splits a single batched Pytree into a list of unbatched Pytrees,
    using the 'original_tree' structure to ensure correct unbatched leaf shapes.

    Args:
        batched: A Pytree where every leaf array has shape (B, ...).
        original: A Pytree with the *expected* leaf shapes after unbatching.

    Returns:
        A list of Pytrees, each corresponding to one element from the batch (B).
    """
    # 1. Separate the structure (aux_data) from the data (batched_leaves)
    batched_leaves, tree_structure = tree.flatten(batched)

    if not batched_leaves:
        # Handle empty or purely structural trees
        return [batched]

    # Get the target unbatched leaf shapes from the original tree
    original_leaves, original_structure = tree.flatten(original)

    # Input validation: Ensure the batched and original tree structures match
    if tree_structure != original_structure:
        raise ValueError("Structure of batched_tree must match original_tree.")

    # Extract target shapes
    target_shapes = [
        leaf.shape if isinstance(leaf, jnp.ndarray) else None for leaf in original_leaves
    ]

    # 2. Determine the batch size (B) from the leading dimension of the first leaf
    try:
        batch_size = batched_leaves[0].shape[0]
        if batch_size == 0:
            return []

    except IndexError:
        # Handle case where the leaf is a scalar JAX array (shape ())
        return [batched]

    # Input validation: Ensure all leaves have the same batch size
    for leaf in batched_leaves:
        if leaf.shape[0] != batch_size:
            raise ValueError(
                f"All leaves must have the same leading batch dimension. "
                f"Found shapes: {leaf.shape} and {batched_leaves[0].shape}"
            )

    # 3. Transpose/Split the leaves list by the batch dimension (B)
    # The inner function splits one array (leaf) into 'B' components (each of shape (1, ...)).
    split_leaves = [jnp.split(leaf, batch_size, axis=0) for leaf in batched_leaves]

    # 'zip' now groups the k-th elements from all split lists together
    unbatched_leaves_list = zip(*split_leaves)

    # 4. Reconstruct the Pytree for each set of unbatched leaves
    unbatched_trees = []
    for leaves_for_batch_k in unbatched_leaves_list:

        reshaped_leaves = []
        for leaf_idx, split_leaf in enumerate(leaves_for_batch_k):

            # The split_leaf is guaranteed to have shape (1, D, W, H) or (1,)
            target_shape = target_shapes[leaf_idx]

            # Use jnp.reshape to force the leaf back to its original shape.
            # This correctly handles the transformation:
            # From (1, D, W, H) -> to (D, W, H)
            # From (1,) -> to ()
            # From (1, 1) -> to (1,)
            # If target_shape is None then we have a scalar (not an array)
            reshaped_leaf = (
                jnp.reshape(split_leaf, target_shape) if target_shape is not None else split_leaf[0]
            )
            reshaped_leaves.append(reshaped_leaf)

        # Recreate the Pytree structure
        unbatched_tree = tree.unflatten(tree_structure, reshaped_leaves)
        unbatched_trees.append(unbatched_tree)

    return unbatched_trees



"""Stage for evaluating dataset statistics"""

from typing import TYPE_CHECKING, TypeVar

import beartype
from flax import nnx
import jaxtyping as jt
from typing_extensions import override

from reax import metrics

from . import stages
from .. import exceptions

if TYPE_CHECKING:
    import reax


__all__ = ("EvaluateStats",)

OutT = TypeVar("OutT")


class EvaluateStats(stages.EpochStage):
    """A stage that can be used to evaluate statistics (in the form of metrics) on a dataset"""

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        stats: "reax.Metric | Sequence[reax.Metric] | dict[str, reax.Metric]",
        datamanager: "reax.data.DataSourceManager",
        engine: "reax.Engine | None",
        *,
        rngs: nnx.Rngs,
        dataset_name: str = "train",
        fast_dev_run: bool | int = False,
        limit_batches: int | float | None = None,
        ignore_missing=False,
    ):
        """Init function."""
        super().__init__(
            "stats",
            None,
            datamanager,
            engine,
            rngs=rngs,
            dataloader_name=dataset_name,
            limit_batches=limit_batches,
            fast_dev_run=fast_dev_run,
        )

        # Params
        self._stats = metrics.MetricCollection(stats)
        self._ignore_missing = ignore_missing

    @override
    def _step(self) -> None:
        """Step function."""
        # Calculate and log all the stats
        for name, stat in self._stats.items():
            try:
                if isinstance(self.batch, tuple):
                    self.log(
                        name, stat.create(*self.batch), on_step=False, on_epoch=True, logger=True
                    )
                else:
                    self.log(
                        name, stat.create(self.batch), on_step=False, on_epoch=True, logger=True
                    )
            except exceptions.DataNotFound:
                if not self._ignore_missing:
                    raise



import abc
from collections.abc import Callable, Iterable, Sequence
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    import reax

__all__ = "Sampler", "DataLoader", "Dataset", "CollateFn"

_T_co = TypeVar("_T_co", covariant=True)
U = TypeVar("U")
IdxT = TypeVar("IdxT")

Dataset = Iterable[_T_co] | Sequence[_T_co]
Sampler = Iterable[IdxT]
CollateFn = Callable[[Sequence[_T_co]], U]


class DataLoader(Generic[_T_co, U], Iterable[U], abc.ABC):
    def __len__(self) -> int:
        return len(self.sampler)

    @property
    @abc.abstractmethod
    def dataset(self) -> "reax.data.Dataset[_T_co]":
        """The dataset being loaded."""

    @property
    @abc.abstractmethod
    def sampler(self) -> "reax.data.Sampler":
        """Access the index sampler used by the dataloader"""

    @abc.abstractmethod
    def with_new_sampler(self, sampler: "reax.data.Sampler") -> "DataLoader[_T_co, U]":
        """Recreate the loader with the given index sampler"""



import logging

from typing_extensions import override

from . import _metric as metric_
from . import collections
from .. import plugins
from ..utils import containers

__all__ = ("Registry", "get_registry", "set_registry", "build_collection")


_LOGGER = logging.getLogger(__name__)


class Registry(containers.Registry[metric_.Metric]):
    def __getitem__(self, key: str) -> metric_.Metric:
        try:
            return self._registry[key]
        except KeyError as exc:
            raise KeyError(f"Metric not found: {key}") from exc

    @override
    def register(self, key: str, obj: metric_.Metric | type[metric_.Metric]):
        """Register function."""
        try:
            if issubclass(obj, metric_.Metric):
                # Try to instantiate it
                obj = obj()
        except TypeError:
            pass

        if not isinstance(obj, metric_.Metric):
            raise ValueError(
                f"metric must be a subclass or instance of `reax.Metric`, got {type(obj).__name__}"
            )

        self._registry[key] = obj.empty()


# Helpers to make it easy to choose a metric using a string
_registry: Registry | None = None


def get_registry() -> Registry:
    """Get registry."""
    global _registry  # pylint: disable=global-statement,global-variable-not-assigned # noqa: F824
    if _registry is None:
        _registry = Registry()
        for name, metric in plugins.get_metrics().items():
            try:
                _registry.register(name, metric)
            except ValueError as exc:
                _LOGGER.warning("Failed to load metric %s: %s", name, exc)

    return _registry


def set_registry(registry: Registry) -> None:
    """Set registry."""
    global _registry  # pylint: disable=global-statement
    _registry = registry


def get(name: str) -> metric_.Metric:
    """Convenience method to get a metric using its name."""
    return get_registry()[name]


def build_collection(items: str | dict | list) -> collections.MetricCollection:
    return collections.MetricCollection(_get_metrics(items))


def _get_metrics(
    items: metric_.Metric | str | dict | list,
) -> metric_.Metric | list[metric_.Metric] | dict[str, metric_.Metric]:
    reg = get_registry()

    if isinstance(items, metric_.Metric):
        return items

    if isinstance(items, str):
        return reg[items]

    if isinstance(items, dict):
        return {name: _get_metrics(value) for name, value in items.items()}

    if isinstance(items, list):
        return [_get_metrics(entry) for entry in items]

    raise TypeError(f"Unknown metrics type: {type(items).__name__}")



from collections.abc import Callable, Generator, Sequence
import contextlib
import functools
import logging
import os
import signal
import sys
from typing import TYPE_CHECKING, Any, Final, TypeVar
import weakref

import beartype
from flax import nnx
import fsspec
import jaxtyping as jt
from lightning_utilities.core import rank_zero
from typing_extensions import override

from . import _checkpointing, _deprecated, _logger_connector
from .. import _engine as engine_
from .. import data, exceptions, hooks, keys
from .. import listeners as listeners_
from .. import loggers as loggers_
from .. import modules, stages, strategies
from ..utils import events

if TYPE_CHECKING:
    import reax

StageT = TypeVar("StageT", bound=stages.Stage)

_LOGGER = logging.getLogger(__name__)

__all__ = ("Trainer",)


class Trainer(stages.StageListener, _deprecated.TrainerDeprecatedMixin):
    # pylint: disable=too-many-public-methods

    _engine = None

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        *,
        accelerator: str = "auto",
        strategy: "str | reax.Strategy" = "auto",
        devices: list[int] | str | int = "auto",
        precision=None,
        logger: "reax.Logger | Iterable[reax.Logger] | bool | None" = True,
        listeners: "list[reax.TrainerListener] | reax.TrainerListener | None" = None,
        log_every_n_steps: int = 50,
        enable_checkpointing: bool | None = True,
        enable_progress_bar: bool = True,
        enable_model_summary: bool | None = None,
        deterministic: bool = False,
        rngs: nnx.Rngs = None,
        default_root_dir: "reax.types.Path | None" = None,
        checkpointing: "reax.Checkpointing" = None,
    ):
        """Init function."""
        # Params
        self._log_every_n_steps: Final[int] = log_every_n_steps
        self._default_root_dir = (
            os.fspath(default_root_dir) if default_root_dir is not None else os.getcwd()
        )

        self._automatic_optimization = True
        self._optimizers = []
        self._stage: stages.Stage | None = None

        # State
        self._engine = engine_.Engine(
            accelerator,
            strategy,
            devices,
            precision=precision,
            logger=logger,
            listeners=listeners,
            deterministic=deterministic,
            rngs=rngs,
            default_root_dir=default_root_dir,
        )

        self._current_epoch: int = 0
        self._global_updates: int = 0
        self._checkpointing = (
            checkpointing if checkpointing is not None else _checkpointing.MsgpackCheckpointing()
        )

        # Init
        self._events = events.EventGenerator[hooks.TrainerListener](
            default_args=(weakref.proxy(self),)
        )

        self._logging = _logger_connector.TrainerLogging()

        if isinstance(listeners, hooks.TrainerListener):
            listeners = [listeners]
        if listeners:
            for listener in listeners:
                self._events.add_listener(listener)

        pbar = _init_progress_bar(self._events.find(), enable_progress_bar)
        if pbar is not None:
            self._events.add_listener(pbar)

        if enable_model_summary:
            _LOGGER.warning("`enable_model_summary` is not supported yet, ignoring")

        if self.checkpoint_listeners:
            if not enable_checkpointing:
                raise exceptions.MisconfigurationException(
                    "Trainer was configured with `enable_checkpointing=False`"
                    " but found `ModelCheckpoint` in listeners list."
                )
        elif enable_checkpointing:
            # Create the default checkpointer
            self._events.add_listener(listeners_.ModelCheckpoint())

    def finalize(self):
        """Clean up the trainer.

        After this called this object should no longer be interacted with.
        """
        self._engine.finalize()
        self._engine = None
        self._events = None

    def __del__(self):
        """Del function."""
        if self._engine is not None:
            self.finalize()

    @property
    def engine(self) -> "reax.Engine":
        """The trainer's engine"""
        return self._engine

    @property
    def strategy(self) -> strategies.Strategy:
        """Strategy function."""
        return self._engine.strategy

    @property
    def rngs(self) -> nnx.Rngs:
        return self._engine.rngs

    @property
    def is_global_zero(self) -> bool:
        """Whether this process is the global zero in multi-node training.

        .. code-block:: python

            def training_step(self, batch, batch_idx):
                if self.trainer.is_global_zero:
                    print("in node 0, accelerator 0")
        """
        return self.engine.is_global_zero

    @property
    def current_epoch(self) -> int:
        """Get the current fitting epoch."""
        return self._current_epoch

    @property
    def global_updates(self) -> int:
        """Get the global number of optimizer updates."""
        if self._stage is not None and isinstance(self._stage, stages.Fit):
            # The stage is running, so count add the number of updates it has performed so far
            return self._global_updates + self._stage.updates

        return self._global_updates

    @property
    def optimizers(self) -> list["reax.Optimizer"]:
        """Optimizers function."""
        return self._optimizers

    @optimizers.setter
    def optimizers(self, opts: list["reax.Optimizer"]) -> None:
        """Optimizers function."""
        self._optimizers = opts

    @property
    def should_stop(self) -> bool:
        """Should stop."""
        if self._stage is None:
            return False
        return self._stage.should_stop

    @should_stop.setter
    def should_stop(self, stop: bool):
        """Should stop."""
        if stop:
            self._stage.stop("None")

    @property
    def sanity_checking(self) -> bool:
        """`True` if currently sanity checking, `False` otherwise"""
        return isinstance(self._stage, stages.Fit) and self._stage.sanity_checking

    @property
    def default_root_dir(self) -> str:
        """Get the fallback directory used for loggers and other components when not explicitly
        specified.
        """
        if _is_local_file_protocol(self._default_root_dir):
            return os.path.normpath(os.path.expanduser(self._default_root_dir))

        return self._default_root_dir

    @property
    def checkpointing(self) -> "reax.Checkpointing":
        return self._checkpointing

    # region Listeners

    @property
    def listeners(self) -> "list[reax.TrainerListener]":
        return self._events.find()

    @property
    def early_stopping_listener(self) -> listeners_.EarlyStopping | None:
        """The first :class:`~reax.listeners.early_stopping.EarlyStopping` listener in the
        Trainer.listeners list, or ``None`` if it doesn't exist.
        """
        listeners = self.early_stopping_listeners
        return listeners[0] if len(listeners) > 0 else None

    @property
    def early_stopping_listeners(self) -> list[listeners_.EarlyStopping]:
        """A list of all instances of :class:`~reax.listeners.early_stopping.EarlyStopping` found in
        the Trainer.listeners list.
        """
        return self._events.find(type=listeners_.EarlyStopping)

    @property
    def checkpoint_listeners(self) -> list["reax.listeners.Checkpointer"]:
        """The first :class:`~reax.listeners.model_checkpoint.ModelCheckpoint` listener in the
        Trainer.listeners list, or ``None`` if it doesn't exist.
        """
        return self._events.find(type=listeners_.Checkpointer)

    @property
    def checkpoint_listener(self) -> "reax.listeners.Checkpointer | None":
        """The first :class:`~reax.listeners.model_checkpoint.ModelCheckpoint` listener in the
        Trainer.listeners list, or ``None`` if it doesn't exist.
        """
        listeners = self.checkpoint_listeners
        return listeners[0] if len(listeners) > 0 else None

    @property
    def progress_bar_listeners(self) -> list["reax.listeners.ProgressBar"]:
        """The first :class:`~reax.listeners.ProgressBar` listener in the
        Trainer.listeners list, or ``None`` if it doesn't exist.
        """
        return self._events.find(type=listeners_.ProgressBar)

    @property
    def progress_bar_listener(self) -> "reax.listeners.ProgressBar | None":
        """The first :class:`~reax.listeners.ProgressBar` listener in the
        Trainer.listeners list, or ``None`` if it doesn't exist.
        """
        listeners = self.progress_bar_listeners
        return listeners[0] if len(listeners) > 0 else None

    # end region

    @property
    def logger(self) -> "reax.Logger | None":
        r"""Get the first (and main) logger."""
        return self._engine.logger

    @property
    def loggers(self) -> list["reax.Logger"]:
        """Get all the loggers."""
        return self._engine.loggers

    @loggers.setter
    def loggers(self, loggers: list["reax.Logger"] | None) -> None:
        """Loggers function."""
        self._engine.loggers = loggers

    @property
    def log_dir(self) -> str | None:
        """The directory for the current experiment. Use this to save images to, etc...

        Note:
            You must call this on all processes. Failing to do so will cause your program to
            stall forever.

        .. code-block:: python

            def training_step(self, batch, batch_idx):
                img = ...
                save_img(img, self.trainer.log_dir)
        """
        if len(self.loggers) > 0:
            if not isinstance(self.loggers[0], loggers_.TensorBoardLogger):
                dirpath = self.loggers[0].save_dir
            else:
                dirpath = self.loggers[0].log_dir
        else:
            dirpath = self.default_root_dir

        dirpath = self.engine.broadcast(dirpath)
        return dirpath

    @property
    def stage(self) -> "reax.Stage | None":
        """Get the current stage if there is one running, otherwise None."""
        return self._stage

    @property
    def progress_bar_metrics(self) -> dict:
        """Progress bar metrics."""
        return self._logging.progress_bar_metrics

    @property
    def listener_metrics(self) -> dict:
        """Callback metrics."""
        return self._logging.listener_metrics

    @property
    def logged_metrics(self) -> dict:
        """Logger metrics."""
        return self._logging.logged_metrics

    @property
    def process_index(self) -> int:
        """Global rank."""
        return self._engine.process_index

    @property
    def local_process_index(self) -> int:
        """Local rank."""
        return self._engine.local_process_index

    @property
    def node_rank(self) -> int:
        """Node rank."""
        return self._engine.node_rank

    @property
    def process_count(self) -> int:
        """World size."""
        return self._engine.process_count

    @property
    def train_dataloader(self) -> "reax.data.DataLoader | None":
        """Train dataloader."""
        if self._stage is not None:
            return getattr(self._stage, "train_dataloader", None)
        return None

    @property
    def val_dataloaders(self) -> "reax.data.DataLoader | None":
        """Train dataloader."""
        if self._stage is not None:
            return getattr(self._stage, "val_dataloader", None)
        return None

    @property
    def _module(self) -> "reax.Module":
        if self._stage is None:
            raise AttributeError("There is no stage running, so module is not available")
        if self._stage.module is None:
            raise AttributeError(
                f"The running stage '{self._stage.__class__.__name__}', does not have a module"
            )
        return self._stage.module

    def log(
        # pylint: disable=unused-argument
        self,
        name: str,
        value,
        *,
        prog_bar: bool = False,
        batch_size: int | None = None,
        logger: bool = None,
        on_step=True,
        on_epoch=True,
    ) -> None:
        """Log function."""
        if self._stage is None:
            raise RuntimeError(
                "Logging is only supported during one of the train/validate/test stages. "
                "There is currently no stage running."
            )

        # Let the stage take care of it
        self._stage.log(
            name,
            value,
            prog_bar=prog_bar,
            logger=logger,
            batch_size=batch_size,
            on_step=on_step,
            on_epoch=on_epoch,
        )

    # region Stages

    def fit(
        self,
        module: modules.Module,
        train_dataloaders: "reax.DataLoader | None" = None,
        val_dataloaders: "reax.DataLoader | None" = None,
        *,
        datamodule: "reax.DataModule | None" = None,
        ckpt_path: "reax.types.Path | None" = None,
        fast_dev_run: bool | int = False,
        max_epochs: int | None = 1_000,
        min_epochs: int = 0,
        min_updates: int = 0,
        max_updates: int | float = None,
        limit_train_batches: int | float | None = 1.0,
        accumulate_grad_batches: int = 1,
        limit_val_batches: int | float | None = 1.0,
        val_check_interval: int | float | None = 1.0,
        check_val_every_n_epoch: int = 1,
        num_sanity_val_steps: int | None = 2,
        reload_dataloaders_every_n_epochs: int = 0,
    ) -> "reax.stages.Fit":
        """Fit function.

        Args:
            ckpt_path: Path/URL of the checkpoint from which training is
                resumed. Could also be one of two special keywords
                ``"last"`` and ``"hpc"``. If there is no checkpoint file
                at the path, an exception is raised.
        """
        if max_updates is not None and max_updates < 0:
            raise ValueError("`max_updates` must be a non-negative integer")
        if max_epochs is not None and max_epochs < 0:
            raise ValueError(f"`max_epochs` must be a non-negative integer or {keys.NO_LIMIT}")

        if fast_dev_run:
            num_batches = 1 if fast_dev_run is True else fast_dev_run
            rank_zero.rank_zero_info(
                f"Running in `fast_dev_run` mode: will run the requested loop using {num_batches} "
                f"batch(es). Logging and checkpointing is suppressed."
            )

        if ckpt_path:
            self._load_module_params(module, ckpt_path)

        datamanager = self._create_data_manager(
            module=module,
            datamodule=datamodule,
            train=train_dataloaders,
            val=val_dataloaders,
        )
        fit = stages.Fit(
            module,
            datamanager,
            self.optimizers,
            self.engine,
            rngs=self.rngs,
            fast_dev_run=fast_dev_run,
            max_epochs=max_epochs,
            min_epochs=min_epochs,
            min_updates=min_updates,
            max_updates=max_updates,
            limit_train_batches=limit_train_batches,
            accumulate_grad_batches=accumulate_grad_batches,
            limit_val_batches=limit_val_batches,
            val_check_interval=val_check_interval,
            check_val_every_n_epoch=check_val_every_n_epoch,
            num_sanity_val_steps=num_sanity_val_steps,
            reload_dataloaders_every_n_epochs=reload_dataloaders_every_n_epochs,
        )
        self._run_stage(fit)
        # Update state variables
        self._global_updates += fit.updates

        return fit

    @jt.jaxtyped(typechecker=beartype.beartype)
    def validate(
        self,
        module: modules.Module,
        dataloaders=None,
        datamodule=None,
        ckpt_path: "reax.types.Path | None" = None,
        limit_batches: int | type(keys.NO_LIMIT) = keys.NO_LIMIT,
    ) -> "reax.stages.Validate":
        """Validate function."""
        if ckpt_path:
            self._load_module_params(module, ckpt_path)

        datamanager = self._create_data_manager(module, datamodule, val=dataloaders)
        validate = stages.Validate(module, datamanager, self.engine, limit_batches=limit_batches)
        self._run_stage(validate)

        return validate

    @jt.jaxtyped(typechecker=beartype.beartype)
    def test(
        self,
        module: modules.Module,
        dataloaders=None,
        datamodule=None,
        ckpt_path: "reax.types.Path | None" = None,
        fast_dev_run: bool | int = False,
        limit_batches: int | float | None = 1.0,
    ) -> "reax.stages.Test":
        """Test function."""
        if ckpt_path:
            self._load_module_params(module, ckpt_path)

        datamanager = self._create_data_manager(module, datamodule, test=dataloaders)
        test = stages.Test(
            module,
            datamanager,
            self.engine,
            rngs=self.rngs,
            fast_dev_run=fast_dev_run,
            limit_batches=limit_batches,
        )
        self._run_stage(test)

        return test

    @jt.jaxtyped(typechecker=beartype.beartype)
    def predict(
        self,
        module: modules.Module,
        dataloaders: "reax.DataLoader | None" = None,
        datamodule: "reax.DataModule | None" = None,
        ckpt_path: "reax.types.Path | None" = None,
        keep_predictions: bool | None = None,
        fast_dev_run: bool | int = False,
        limit_batches: int | float = keys.NO_LIMIT,
    ) -> "reax.stages.Predict":
        r"""Run inference on the data.

        Logging is disabled in the predict hooks.
        """
        if ckpt_path:
            self._load_module_params(module, ckpt_path)

        if fast_dev_run:
            limit_batches = 1

        if keep_predictions is None:
            keep_predictions = True

        datamanager = self._create_data_manager(module, datamodule, predict=dataloaders)
        predict = stages.Predict(
            module,
            datamanager,
            self.engine,
            limit_batches=limit_batches,
            keep_predictions=keep_predictions,
        )
        self._run_stage(predict)

        return predict

    @jt.jaxtyped(typechecker=beartype.beartype)
    def eval_stats(
        self,
        stats: "reax.Metric | Sequence[reax.Metric] | dict[str, reax.Metric]",
        dataloaders: "reax.DataLoader | None" = None,
        datamodule: "reax.DataModule | None" = None,
        dataset_name: str = "train",
        fast_dev_run: bool | int = False,
        limit_batches: int | float = keys.NO_LIMIT,
    ) -> "reax.stages.EvaluateStats":
        r"""Evaluate metrics on a dataset to get statistics about it.

        The calculated statistics will be stored on `stage.logged_metrics` where `stage` is the
        return value of this call.
        """
        if fast_dev_run:
            limit_batches = 1

        datamanager = self._create_data_manager(
            datamodule=datamodule, **{dataset_name: dataloaders}
        )
        eval_stats = stages.EvaluateStats(
            stats,
            datamanager,
            self.engine,
            rngs=self.rngs,
            dataset_name=dataset_name,
            limit_batches=limit_batches,
        )
        self._run_stage(eval_stats)

        return eval_stats

    def run(self, stage: StageT) -> StageT:
        return self._run_stage(stage)

    def _run_stage(self, stage: stages.Stage) -> stages.Stage:
        """Run stage."""
        with self.engine.default_device():
            try:
                with self._attach(stage):
                    with stage.events.listen_context(self):
                        stage.run()
            except KeyboardInterrupt as exc:
                rank_zero.rank_zero_info(
                    "\nDetected KeyboardInterrupt, attempting graceful shutdown ..."
                )
                # Disable further Ctrl+C presses while we respond to this one
                signal.signal(signal.SIGINT, signal.SIG_IGN)
                self._interrupt(stage, exc)
                sys.exit(1)
            except BaseException as exc:
                self._interrupt(stage, exc)
                raise

            else:
                for logger in self.loggers:
                    logger.finalize("success")

        return stage

    def _interrupt(self, stage: "reax.Stage", exception: BaseException):
        # Inform our listeners of the exception
        self._events.fire_event(hooks.TrainerListener.on_exception, stage, exception)

        # TODO: self._strategy.on_exception(exc)
        for logger in self.loggers:
            logger.finalize("failed")

    # endregion

    @contextlib.contextmanager
    def _attach(self, stage: stages.Stage):
        """Attach function."""
        self._stage = stage
        if stage.module is not None:
            listeners = stage.module.configure_listeners()
        else:
            listeners = []

        module_trainer_ctx = contextlib.nullcontext if stage.module is None else stage.module.attach

        try:
            with module_trainer_ctx(weakref.proxy(self)), self._attach_model_listeners(listeners):
                yield
        finally:
            self._stage = None

    @contextlib.contextmanager
    def _attach_model_listeners(
        self, listeners: "reax.TrainerListener | Sequence[reax.TrainerListener]"
    ) -> Generator[None, Any, None]:
        """Attaches the listeners defined in the model.

        If a listener returned by the model's configure_listener method has the same type as one or
        several listeners already present in the trainer listeners list, it will replace them.
        In addition, all :class:`~lightning.pytorch.listeners.model_checkpoint.ModelCheckpoint`
        listeners will be pushed to the end of the list, ensuring they run last.
        """
        if listeners:
            listeners = [listeners] if not isinstance(listeners, Sequence) else listeners
            model_listener_types = {type(c) for c in listeners}
            trainer_listener_types = {type(c) for c in self.listeners}
            # edge case: if an unmodified listener was added, the logic below would filter it
            trainer_listener_types.discard(hooks.TrainerListener)
            # exclude trainer listeners of the same class or subclass
            override_types = set()
            for model_cb in model_listener_types:
                for trainer_cb in trainer_listener_types:
                    if issubclass(model_cb, trainer_cb):
                        override_types.add(trainer_cb)
                        break
            if override_types:
                rank_zero.rank_zero_info(
                    "The following listeners returned in `Module.configure_listeners` will override"
                    " existing listeners passed to Trainer:"
                    f" {', '.join(sorted(t.__name__ for t in override_types))}"
                )
            # remove all listeners with a type that occurs in model listeners
            all_listeners = [
                listener for listener in self.listeners if type(listener) not in override_types
            ]
            all_listeners.extend(listeners)
            all_listeners = _reorder_listeners(all_listeners)
        else:
            all_listeners = None

        original_events = self._events
        try:
            if listeners:
                self._events = events.EventGenerator[hooks.TrainerListener](
                    default_args=(weakref.proxy(self),)
                )
                for listener in all_listeners:
                    self._events.add_listener(listener)

            yield
        finally:
            if listeners:
                self._events = original_events

    @override
    def on_stage_start(self, stage: "stages.Stage", /) -> None:
        """The stage is starting."""
        if not self._optimizers and isinstance(stage, stages.Train):
            self._optimizers = stage.optimizers

        if stage.is_root:
            self._logging.reset_metrics()
            self._events.fire_event(hooks.TrainerListener.setup, stage)

        self._events.fire_event(hooks.TrainerListener.on_stage_start, stage)
        self._fire_stage_event(stage, stages.StageListener.on_stage_start)

    @override
    def on_stage_started(self, stage: "reax.Stage", /):
        if isinstance(stage, stages.EpochStage):
            self._events.fire_event(hooks.TrainerListener.on_epoch_start, stage)
        self._fire_stage_event(stage, stages.StageListener.on_stage_started)

    @override
    def on_stage_iter_starting(self, stage: "stages.Stage", step: int, /):
        """On stage iter starting."""
        self._events.fire_event(hooks.TrainerListener.on_stage_iter_start, stage, step)
        if isinstance(stage, stages.EpochStage):
            self._fire_stage_event(
                stage, stages.StageListener.on_stage_iter_starting, stage.batch, step
            )

    @override
    def on_stage_iter_ended(self, stage: "stages.Stage", step: int, outputs: Any, /):
        """On stage iter ending."""
        if isinstance(stage, stages.EpochStage):
            self._logging.update(stage)
            logging_metrics = {"epoch": self.current_epoch, "stage": stage.name}
            logging_metrics.update(stage.logged_metrics)
            for logger in self.loggers:
                logger.log_metrics(metrics=logging_metrics, step=self.global_updates - 1)
                logger.save()

            self._events.fire_event(hooks.TrainerListener.on_batch_end, stage, step, outputs)

            self._fire_stage_event(
                stage, stages.StageListener.on_stage_iter_ending, outputs, stage.batch, step
            )

        self._events.fire_event(hooks.TrainerListener.on_stage_iter_end, stage, step, outputs)

        if isinstance(stage, stages.Fit):
            # Keep track of the number of completed training epochs
            self._current_epoch += 1

    @override
    def on_stage_end(self, stage: "reax.Stage", /):
        if isinstance(stage, stages.EpochStage):
            self._logging.update(stage)
            self._events.fire_event(
                hooks.TrainerListener.on_epoch_end, stage, stage.listener_metrics
            )

            metrics = stage.results
            if metrics[keys.LOG]:
                logging_metrics = {"epoch": self.current_epoch, "stage": stage.name}
                logging_metrics.update(self.logged_metrics)
                for logger in self.loggers:
                    logger.log_metrics(metrics=logging_metrics, step=self.global_updates - 1)
                    logger.save()

        self._fire_stage_event(stage, stages.StageListener.on_stage_end)

    @override
    def on_stage_ended(self, stage: "reax.Stage", /):
        self._events.fire_event(hooks.TrainerListener.on_stage_end, stage)
        self._fire_stage_event(stage, stages.StageListener.on_stage_ended)
        # Finally, call teardown
        if stage.is_root:
            self._events.fire_event(hooks.TrainerListener.teardown, stage)

    # region Checkpointing

    def save_checkpoint(self, filepath: "reax.types.Path", weights_only: bool = True):
        """For now, we just save the model weights.

        The user has to store the model definition themselves.
        """
        ckpt = _checkpointing.dump_checkpoint(self._module, weights_only=weights_only)
        self._checkpointing.save(ckpt, filepath)

    def _get_checkpoint_path(self) -> str | None:
        """Get checkpoint path."""
        checkpointers = self.checkpoint_listeners
        if not checkpointers:
            return None

        if len(checkpointers) > 1:
            fn = self._get_checkpoint_path
            rank_zero.rank_zero_warn(
                f'`.{fn}(ckpt_path="best")` found Trainer with multiple `ModelCheckpoint`'
                " listeners. The best checkpoint path from first checkpoint listener will be used."
            )

        checkpointer = checkpointers[0]
        return getattr(checkpointer, "best_model_path", None)

    # endregion

    def _fire_stage_event(self, stage, hook: Callable, *args):
        event = hook_map(stage).get(hook)
        if event is not None:
            self._events.fire_event(event, stage, *args)

    def _load_module_params(self, module: "reax.Module", filename: str):
        ckpt = self._checkpointing.load(filename)
        module.set_parameters(ckpt[_checkpointing.PARAMS])

    def _create_data_manager(
        self,
        module: "reax.data.DataSource" = None,
        datamodule: "reax.data.DataModule" = None,
        **loaders,
    ) -> data.DataSourceManager:
        return data.create_manager(
            module=module, datamodule=datamodule, engine=self._engine, **loaders
        )


@jt.jaxtyped(typechecker=beartype.beartype)
def _init_loggers(
    logger: "reax.Logger | Iterable[reax.Logger] | bool | None", default_root_dir: str
) -> list["reax.Logger"]:
    """Init loggers."""
    if isinstance(logger, loggers_.Logger):
        return [logger]

    if isinstance(logger, (bool, type(None))):
        if logger:
            return [loggers_.CsvLogger(default_root_dir)]

        return []

    return list(logger)


def _init_progress_bar(
    listeners: list[hooks.TrainerListener], enable_progress_bar: bool = True
) -> listeners_.ProgressBar | None:
    progress_bars = [
        listener for listener in listeners if isinstance(listener, listeners_.progress.ProgressBar)
    ]
    if len(progress_bars) > 1:
        raise exceptions.MisconfigurationException(
            "You added multiple progress bar listeners to the Trainer, but currently only one"
            " progress bar is supported."
        )
    if len(progress_bars) == 1:
        # the user specified the progress bar in the listeners list
        # so the trainer doesn't need to provide a default one
        if enable_progress_bar:
            return None

        # otherwise the user specified a progress bar listener but also
        # elected to disable the progress bar with the trainer flag
        progress_bar_listener = progress_bars[0]
        raise exceptions.MisconfigurationException(
            "Trainer was configured with `enable_progress_bar=False`"
            f" but found `{progress_bar_listener.__class__.__name__}` in listeners list."
        )

    if enable_progress_bar:
        return listeners_.progress.TqdmProgressBar()

    return None


def _is_local_file_protocol(path: "reax.types.Path") -> bool:
    """Is local file protocol."""
    return fsspec.utils.get_protocol(str(path)) == "file"


@functools.singledispatch
def hook_map(_stage) -> dict[Callable, Callable]:
    """Hook map."""
    return dict()


@hook_map.register
def _(_stage: stages.Train) -> dict[Callable, Callable]:
    """Function."""
    return {
        stages.StageListener.on_stage_start: hooks.TrainerListener.on_train_start,
        stages.StageListener.on_stage_started: hooks.TrainerListener.on_train_epoch_start,
        stages.StageListener.on_stage_iter_starting: hooks.TrainerListener.on_train_batch_start,
        stages.StageListener.on_stage_iter_ending: hooks.TrainerListener.on_train_batch_end,
        stages.StageListener.on_stage_end: hooks.TrainerListener.on_train_epoch_end,
        stages.StageListener.on_stage_ended: hooks.TrainerListener.on_train_end,
    }


@hook_map.register
def _(_stage: stages.Validate) -> dict[Callable, Callable]:
    """Function."""
    return {
        stages.StageListener.on_stage_start: hooks.TrainerListener.on_validation_start,
        stages.StageListener.on_stage_started: hooks.TrainerListener.on_validation_epoch_start,
        # pylint: disable=line-too-long
        stages.StageListener.on_stage_iter_starting: hooks.TrainerListener.on_validation_batch_start,
        stages.StageListener.on_stage_iter_ending: hooks.TrainerListener.on_validation_batch_end,
        stages.StageListener.on_stage_end: hooks.TrainerListener.on_validation_epoch_end,
        stages.StageListener.on_stage_ended: hooks.TrainerListener.on_validation_end,
    }


@hook_map.register
def _(_stage: stages.Test) -> dict[Callable, Callable]:
    """Function."""
    return {
        stages.StageListener.on_stage_start: hooks.TrainerListener.on_test_start,
        stages.StageListener.on_stage_started: hooks.TrainerListener.on_test_epoch_start,
        stages.StageListener.on_stage_iter_starting: hooks.TrainerListener.on_test_batch_start,
        stages.StageListener.on_stage_iter_ending: hooks.TrainerListener.on_test_batch_end,
        stages.StageListener.on_stage_end: hooks.TrainerListener.on_test_epoch_end,
        stages.StageListener.on_stage_ended: hooks.TrainerListener.on_test_end,
    }


@hook_map.register
def _(_stage: stages.Predict) -> dict[Callable, Callable]:
    """Function."""
    return {
        stages.StageListener.on_stage_start: hooks.TrainerListener.on_predict_start,
        stages.StageListener.on_stage_started: hooks.TrainerListener.on_predict_epoch_start,
        stages.StageListener.on_stage_iter_starting: hooks.TrainerListener.on_predict_batch_start,
        stages.StageListener.on_stage_iter_ending: hooks.TrainerListener.on_predict_batch_end,
        stages.StageListener.on_stage_end: hooks.TrainerListener.on_predict_epoch_end,
        stages.StageListener.on_stage_ended: hooks.TrainerListener.on_predict_end,
    }


@hook_map.register
def _(_stage: stages.Fit) -> dict[Callable, Callable]:
    """Function."""
    return {
        stages.StageListener.on_stage_start: hooks.TrainerListener.on_fit_start,
        stages.StageListener.on_stage_end: hooks.TrainerListener.on_fit_end,
    }


def _reorder_listeners(listeners: list["reax.TrainerListener"]) -> list["reax.TrainerListener"]:
    """Moves all the tuner specific listeners at the beginning of the list and all the
    `ModelCheckpoint` listeners to the end of the list. The sequential order within the group of
    checkpoint listeners is preserved, as well as the order of all other listeners.

    Args:
        listeners: A list of listeners.

    Return:
        A new list in which the first elements are tuner specific listeners and last elements are
        ModelCheckpoints if there were any present in the input.

    """
    tuner_listeners: list["reax.TrainerListener"] = []
    other_listeners: list["reax.TrainerListener"] = []
    checkpoint_listeners: list["reax.TrainerListener"] = []

    for cb in listeners:
        # if isinstance(cb, (BatchSizeFinder, LearningRateFinder)):
        #     tuner_listeners.append(cb)
        if isinstance(cb, listeners_.Checkpointer):
            checkpoint_listeners.append(cb)
        else:
            other_listeners.append(cb)

    return tuner_listeners + other_listeners + checkpoint_listeners



# Copyright (C) 2025  Martin Uhrin
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Most of this file is covered by the following license.  To find what has been modified you
# can perform a diff with the file at:
# https://github.com/Lightning-AI/pytorch-lightning/blob/0324a20f00235c7a10a235a44326811ba42b6ae4/src/lightning/pytorch/utilities/grads.py
#
# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities to describe gradients."""

from typing import Any

import jax.numpy as jnp
from pytray import tree

__all__ = ("grad_norm",)


def grad_norm(
    grads: dict[str, Any], norm_type: float | int | str, group_separator: str = "/"
) -> dict[str, float]:
    """Compute each parameter's gradient's norm and their overall norm.

    The overall norm is computed over all gradients together, as if they
    were concatenated into a single vector.

    Args:
        grads: the gradients PyTree, typically coming from a ``.grad()`` call
        norm_type: The type of the used p-norm, cast to float if necessary.
            Can be ``'inf'`` for infinity norm.
        group_separator: The separator string used by the logger to group
            the gradients norms in their own subfolder instead of the logs one.

    Return:
        norms: The dictionary of p-norms of each parameter's gradient and
            a special entry for the total p-norm of the gradients viewed
            as a single vector.

    """
    norm_type = float(norm_type)
    if norm_type <= 0:
        raise ValueError(
            f"`norm_type` must be a positive number or 'inf' (infinity norm). Got {norm_type}"
        )

    def path_name(path) -> str:
        """Given a tree path this will return a string version of the path"""
        path_str = tree.path_to_string(path).replace(" ", "_")
        return f"grad_{norm_type}_norm{group_separator}{path_str}"

    norms = {
        path_name(path): jnp.linalg.norm(value.flatten(), ord=norm_type)
        for path, value in tree.flatten(grads).items()
        if value is not None
    }

    if norms:
        total_norm = jnp.linalg.norm(jnp.array(list(norms.values())), ord=norm_type)
        norms[f"grad_{norm_type}_norm_total"] = total_norm
    return norms



import subprocess
import sys


def probe_local_device_count(platform: str) -> int:
    """Use a subprocess to import JAX and ask it the number of local devices.  This is necessary
    if we want to avoid calling any jax functions in this process which would already set in stone
    the device setup
    """
    if platform == "auto":
        platform = "None"
    else:
        platform = f"'{platform}'"

    code = [
        "import jax",
        f"jax.config.update('jax_platforms', {platform})",
        "print(jax.local_device_count())",
    ]

    code = ";".join(code)
    result = subprocess.check_output([sys.executable, "-c", code])  # nosec
    return int(result.decode().strip())



import os
import sys
from typing import TYPE_CHECKING, Any

from lightning_utilities.core import rank_zero
import tqdm.auto as tqdm
from typing_extensions import override

from . import progress_bar
from .. import utils

if TYPE_CHECKING:
    import reax

__all__ = ("TqdmProgressBar",)


class TqdmProgressBar(progress_bar.ProgressBar):
    # pylint: disable=too-many-public-methods
    BAR_FORMAT = (
        "{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_noinv_fmt}{postfix}]"
    )

    def __init__(self, refresh_rate: int = 1, process_position: int = 0):
        """Init function."""
        super().__init__()
        self._refresh_rate = self._resolve_refresh_rate(refresh_rate)
        self._process_position = process_position
        self._enabled = True
        self._train_progress_bar: tqdm.tqdm | None = None
        self._val_progress_bar: tqdm.tqdm | None = None
        self._test_progress_bar: tqdm.tqdm | None = None
        self._predict_progress_bar: tqdm.tqdm | None = None

    @property
    def train_progress_bar(self) -> tqdm.tqdm:
        """Train progress bar."""
        if self._train_progress_bar is None:
            raise TypeError(
                f"The `{self.__class__.__name__}._train_progress_bar` reference has not been set "
                f"yet."
            )
        return self._train_progress_bar

    @train_progress_bar.setter
    def train_progress_bar(
        # pylint: disable=disallowed-name
        self,
        bar: tqdm.tqdm,
    ) -> None:
        """Train progress bar."""
        self._train_progress_bar = bar

    @property
    def val_progress_bar(self) -> tqdm.tqdm:
        """Val progress bar."""
        if self._val_progress_bar is None:
            raise TypeError(
                f"The `{self.__class__.__name__}._val_progress_bar` reference has not been set yet."
            )
        return self._val_progress_bar

    @val_progress_bar.setter
    def val_progress_bar(
        # pylint: disable=disallowed-name
        self,
        bar: tqdm.tqdm,
    ) -> None:
        """Val progress bar."""
        self._val_progress_bar = bar

    @property
    def test_progress_bar(self) -> tqdm.tqdm:
        """Test progress bar."""
        if self._test_progress_bar is None:
            raise TypeError(
                f"The `{self.__class__.__name__}._test_progress_bar` reference has not been set "
                f"yet."
            )
        return self._test_progress_bar

    @test_progress_bar.setter
    def test_progress_bar(
        # pylint: disable=disallowed-name
        self,
        bar: tqdm.tqdm,
    ) -> None:
        """Test progress bar."""
        self._test_progress_bar = bar

    @property
    def predict_progress_bar(self) -> tqdm.tqdm:
        """Predict progress bar."""
        if self._predict_progress_bar is None:
            raise TypeError(
                f"The `{self.__class__.__name__}._predict_progress_bar` reference has not been set "
                f"yet."
            )
        return self._predict_progress_bar

    @predict_progress_bar.setter
    def predict_progress_bar(
        # pylint: disable=disallowed-name
        self,
        bar: tqdm.tqdm,
    ) -> None:
        """Predict progress bar."""
        self._predict_progress_bar = bar

    @property
    def refresh_rate(self) -> int:
        """Refresh rate."""
        return self._refresh_rate

    @property
    def process_position(self) -> int:
        """Process position."""
        return self._process_position

    @property
    def is_enabled(self) -> bool:
        """Is enabled."""
        return self._enabled and self.refresh_rate > 0

    @property
    def is_disabled(self) -> bool:
        """Is disabled."""
        return not self.is_enabled

    @override
    def disable(self) -> None:
        """Disable function."""
        self._enabled = False

    @override
    def enable(self) -> None:
        """Enable function."""
        self._enabled = True

    def init_train_tqdm(self, stage: "reax.stages.Fit") -> tqdm.tqdm:
        """Override this to customize the tqdm bar for training."""
        return tqdm.tqdm(
            desc=stage.name,
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=sys.stdout,
            smoothing=0,
            bar_format=self.BAR_FORMAT,
        )

    def init_predict_tqdm(self, stage: "reax.stages.Predict") -> tqdm.tqdm:
        """Override this to customize the tqdm bar for predicting."""
        return tqdm.tqdm(
            desc=stage.name,
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=sys.stdout,
            smoothing=0,
            bar_format=self.BAR_FORMAT,
        )

    def init_validation_tqdm(self, stage: "reax.stages.Validate") -> tqdm.tqdm:
        """Override this to customize the tqdm bar for validation."""
        # Check if there is a 'parent' in which case it will have a progress bar
        has_main_bar = stage.parent is not None
        return tqdm.tqdm(
            desc=stage.name,
            position=(2 * self.process_position + has_main_bar),
            disable=self.is_disabled,
            leave=not has_main_bar,
            dynamic_ncols=True,
            file=sys.stdout,
            bar_format=self.BAR_FORMAT,
        )

    def init_test_tqdm(self) -> tqdm.tqdm:
        """Override this to customize the tqdm bar for testing."""
        return tqdm.tqdm(
            desc="Testing",
            position=(2 * self.process_position),
            disable=self.is_disabled,
            leave=True,
            dynamic_ncols=True,
            file=sys.stdout,
            bar_format=self.BAR_FORMAT,
        )

    @override
    def on_fit_start(self, _trainer: "reax.Trainer", stage: "reax.stages.Fit", /) -> None:
        """On fit start."""
        self.train_progress_bar = self.init_train_tqdm(stage)

    @override
    def on_train_epoch_start(self, trainer: "reax.Trainer", stage: "reax.stages.Train", /) -> None:
        """On train epoch start."""
        self.train_progress_bar.reset(total=utils.convert_inf(stage.max_batches))
        self.train_progress_bar.initial = 0
        self.train_progress_bar.set_description(f"Epoch {trainer.current_epoch}")

    @override
    def on_train_batch_end(
        self,
        trainer: "reax.Trainer",
        _stage: "reax.stages.Train",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """On train batch end."""
        n = batch_idx + 1
        if self._should_update(n, self.train_progress_bar.total):
            _update_n(self.train_progress_bar, n)
            self.train_progress_bar.set_postfix(trainer.progress_bar_metrics)

    @override
    def on_train_epoch_end(self, trainer: "reax.Trainer", _stage: "reax.stages.Train", /) -> None:
        """On train epoch end."""
        if not self.train_progress_bar.disable:
            self.train_progress_bar.set_postfix(trainer.progress_bar_metrics)

    @override
    def on_fit_end(self, _trainer: "reax.Trainer", _stage: "reax.stages.Fit", /) -> None:
        """On fit end."""
        self.train_progress_bar.close()

    @override
    def on_validation_epoch_start(
        self, _trainer: "reax.Trainer", stage: "reax.stages.Validate", /
    ) -> None:
        """On validation epoch start."""
        self.val_progress_bar = self.init_validation_tqdm(stage)
        self.val_progress_bar.reset(total=utils.convert_inf(stage.max_batches))
        self.val_progress_bar.initial = 0
        self.val_progress_bar.set_description(stage.name)

    @override
    def on_validation_batch_end(
        self,
        _trainer: "reax.Trainer",
        _stage: "reax.Stage",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """On validation batch end."""
        n = batch_idx + 1
        if self._should_update(n, self.val_progress_bar.total):
            _update_n(self.val_progress_bar, n)

    @override
    def on_validation_epoch_end(
        self, trainer: "reax.Trainer", stage: "reax.stages.Validate", /
    ) -> None:
        """On validation epoch end."""
        self.val_progress_bar.close()
        if self._train_progress_bar is not None and trainer.stage != stage:
            self.train_progress_bar.set_postfix(trainer.progress_bar_metrics)

    @override
    def on_test_epoch_start(self, _trainer: "reax.Trainer", _stage: "reax.Stage", /) -> None:
        """On test epoch start."""
        self.test_progress_bar = self.init_test_tqdm()

    @override
    def on_test_batch_start(
        self, _trainer: "reax.Trainer", stage: "reax.stages.Test", _batch: Any, _batch_idx: int, /
    ) -> None:
        """On test batch start."""
        self.test_progress_bar.reset(total=utils.convert_inf(stage.max_batches))
        self.test_progress_bar.initial = 0
        self.test_progress_bar.set_description(f"{stage.name}")

    @override
    def on_test_batch_end(
        self,
        _trainer: "reax.Trainer",
        _stage: "reax.Stage",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """On test batch end."""
        n = batch_idx + 1
        if self._should_update(n, self.test_progress_bar.total):
            _update_n(self.test_progress_bar, n)

    @override
    def on_test_epoch_end(self, _trainer: "reax.Trainer", _stage: "reax.Stage", /) -> None:
        """On test epoch end."""
        self.test_progress_bar.close()

    @override
    def on_predict_epoch_start(
        self, _trainer: "reax.Trainer", stage: "reax.stages.Predict", /
    ) -> None:
        """On predict epoch start."""
        self.predict_progress_bar = self.init_predict_tqdm(stage)
        self.predict_progress_bar.reset(total=utils.convert_inf(stage.max_batches))
        self.predict_progress_bar.initial = 0
        self.predict_progress_bar.set_description(stage.name)

    @override
    def on_predict_batch_end(
        self,
        _trainer: "reax.Trainer",
        _stage: "reax.Stage",
        _outputs: Any,
        _batch: Any,
        batch_idx: int,
        /,
    ) -> None:
        """On predict batch end."""
        n = batch_idx + 1
        if self._should_update(n, self.predict_progress_bar.total):
            _update_n(self.predict_progress_bar, n)

    @override
    def on_predict_epoch_end(self, _trainer: "reax.Trainer", _stage: "reax.Stage", /) -> None:
        """On predict epoch end."""
        self.predict_progress_bar.close()

    @override
    def print(self, *args: Any, sep: str = " ", **kwargs: Any) -> None:
        """Print function."""
        active_progress_bar = None

        if self._train_progress_bar is not None and not self.train_progress_bar.disable:
            active_progress_bar = self.train_progress_bar
        elif self._val_progress_bar is not None and not self.val_progress_bar.disable:
            active_progress_bar = self.val_progress_bar
        elif self._test_progress_bar is not None and not self.test_progress_bar.disable:
            active_progress_bar = self.test_progress_bar
        elif self._predict_progress_bar is not None and not self.predict_progress_bar.disable:
            active_progress_bar = self.predict_progress_bar

        if active_progress_bar is not None:
            s = sep.join(map(str, args))
            active_progress_bar.write(s, **kwargs)

    def _should_update(self, current: int, total: int) -> bool:
        """Should update."""
        return self.is_enabled and (current % self.refresh_rate == 0 or current == total)

    @staticmethod
    def _resolve_refresh_rate(refresh_rate: int) -> int:
        """Resolve refresh rate."""
        if os.getenv("COLAB_GPU") and refresh_rate == 1:
            # smaller refresh rate on colab causes crashes, choose a higher value
            rank_zero.rank_zero_debug("Using a higher refresh rate on Colab. Setting it to `20`")
            return 20
        # Support TQDM_MINITERS environment variable, which sets the minimum refresh rate
        if "TQDM_MINITERS" in os.environ:
            return max(int(os.environ["TQDM_MINITERS"]), refresh_rate)
        return refresh_rate


def _update_n(
    # pylint: disable=disallowed-name
    bar: tqdm.tqdm,
    value: int,
) -> None:
    """Update n."""
    if not bar.disable:
        bar.n = value
        bar.refresh()



from . import common, fit, predict, stages, stats, test, train, validation
from .common import *
from .fit import *
from .predict import *
from .stages import *
from .stats import *
from .test import *
from .train import *
from .validation import *

__all__ = (
    common.__all__
    + fit.__all__
    + predict.__all__
    + stages.__all__
    + stats.__all__
    + test.__all__
    + train.__all__
    + validation.__all__
)



import argparse
import contextlib
import copy
import enum
import os
import warnings

import fsspec
from lightning_utilities.core import apply_func
import yaml

try:
    import omegaconf

    _OMEGACONF_AVAILABLE = True
except ImportError:
    _OMEGACONF_AVAILABLE = False

from reax.lightning import rank_zero

from . import types

__all__ = "save_hparams_to_yaml", "load_hparams_from_yaml"


def save_hparams_to_yaml(
    config_yaml: types.Path, hparams: dict | argparse.Namespace, use_omegaconf: bool = True
) -> None:
    """Save the hparams to a yaml file.

    Args:
        config_yaml (typing.Path): The path to save the hparams to.
        hparams (dict | argparse.Namespace): The hparams to save.
        use_omegaconf (bool, optional): If `omegaconf` is available and
            `use_omegaconf=True`, the hparams will be converted to
            `omegaconf.DictConfig` if possible, defaults to True.
    """
    fs = fsspec.url_to_fs(config_yaml)[0]
    if not fs.isdir(os.path.dirname(config_yaml)):
        raise RuntimeError(f"Missing folder: {os.path.dirname(config_yaml)}.")

    # convert Namespace to dict
    if isinstance(hparams, argparse.Namespace):
        hparams = vars(hparams)

    # saving with OmegaConf objects
    if _OMEGACONF_AVAILABLE and use_omegaconf:
        # deepcopy: hparams from user shouldn't be resolved
        hparams = copy.deepcopy(hparams)
        hparams = apply_func.apply_to_collection(
            hparams, omegaconf.DictConfig, omegaconf.OmegaConf.to_container, resolve=True
        )
        with fs.open(config_yaml, "w", encoding="utf-8") as fp:
            try:
                omegaconf.OmegaConf.save(hparams, fp)
                return
            except (omegaconf.UnsupportedValueType, omegaconf.ValidationError):
                pass

    if not isinstance(hparams, dict):
        raise TypeError("hparams must be dictionary")

    hparams_allowed = {}
    # drop parameters which contain some strange datatypes as fsspec
    for key, value in hparams.items():
        try:
            value = value.name if isinstance(value, enum.Enum) else value
            yaml.dump(value)
        except (TypeError, ValueError):
            warnings.warn(
                f"Skipping '{key}' parameter because it is not possible to safely dump to YAML."
            )
            hparams[key] = type(value).__name__
        else:
            hparams_allowed[key] = value

    # saving the standard way
    with fs.open(config_yaml, "w", newline="") as fp:
        yaml.dump(hparams_allowed, fp)


def load_hparams_from_yaml(
    config_yaml: types.Path, use_omegaconf: bool = True
) -> "dict[str, Any] | omegaconf.DictConfig":
    """Load hparams from a file.

    Args:
        config_yaml (typing.Path): Path to the yaml file to be loaded.
        use_omegaconf (bool, optional): If `omegaconf` is available and
            `use_omegaconf=True`, the hparams will be converted to a
            `omegaconf.DictConfig` if possible, defaults to True.
    """
    fs: fsspec.AbstractFileSystem = fsspec.url_to_fs(config_yaml)[0]
    if not fs.exists(config_yaml):
        rank_zero.rank_zero_warn(f"Missing Tags: {config_yaml}.", category=RuntimeWarning)
        return {}

    with fs.open(config_yaml, "r") as fp:
        hparams = yaml.full_load(fp)

    if _OMEGACONF_AVAILABLE and use_omegaconf:
        with contextlib.suppress(omegaconf.UnsupportedValueType, omegaconf.ValidationError):
            return omegaconf.OmegaConf.create(hparams)

    return hparams



from collections.abc import Callable, Iterable, Mapping
import dataclasses
import logging
from typing import Any, TypeVar, cast

import jax
import numpy as np

from reax.utils import containers

from .. import plugins

_LOGGER = logging.getLogger(__name__)

__all__ = "extract_batch_size", "sized_len", "get_registry"

Extractor = Callable[[Any], Iterable[int]]
T = TypeVar("T", bound=type)


class BatchSizer:
    """Tool for extracting batch sizes from dataset."""

    def __init__(self):
        self._registry = containers.TypeRegistry[Extractor]()

    def register(
        self,
        entry_type: T | tuple[type, ...],
        extractor: Callable[[T | tuple[type, ...]], Iterable[int]],
    ):
        """Register function."""
        self._registry.register(entry_type, extractor)

    def extract_batch_size(self, batch) -> Iterable[int]:
        """Try to extract batch size."""
        extractor = self._registry.find(batch)
        if extractor is not None:
            yield from extractor(batch)
        else:
            yield from self._fallback_extractor(batch)

    def _fallback_extractor(self, batch) -> Iterable[int]:
        """Fallback extractor."""
        if isinstance(batch, (Iterable, Mapping)):
            if isinstance(batch, Mapping):
                batch = batch.values()

            for entry in batch:
                yield from self.extract_batch_size(entry)

        elif dataclasses.is_dataclass(batch) and not isinstance(batch, type):
            batch = cast(dataclasses.dataclass, batch)
            for field in dataclasses.fields(batch):
                yield from self.extract_batch_size(getattr(batch, field.name))

        else:
            yield None


_registry = None  # pylint: disable=invalid-name


def _array_batch_size(batch: jax.Array | np.ndarray):
    """Array batch size."""
    if batch.ndim == 0:
        yield 1
    else:
        yield batch.shape[0]


try:
    import torch

    def _tensor_batch_size(batch: "torch.Tensor"):
        """Tensor batch size."""
        if batch.ndim == 0:
            yield 1
        else:
            yield batch.shape[0]

except ImportError:
    torch = None


def get_registry() -> BatchSizer:
    """Get registry."""
    global _registry  # pylint: disable=global-statement
    if _registry is None:
        _registry = BatchSizer()
        _registry.register((jax.Array, np.ndarray), _array_batch_size)
        if torch is not None:
            _registry.register(torch.Tensor, _tensor_batch_size)

        for sizer in plugins.get_batch_sizers():
            _registry.register(*sizer)

    return _registry


def extract_batch_size(batch) -> int:
    """Extract batch size."""
    error_msg = (
        "Could not determine batch size automatically.  You can provide this manually using "
        "`self.log(..., batch_size=size)`"
    )
    batch_size = None
    try:
        for size in get_registry().extract_batch_size(batch):
            if batch_size is None:
                batch_size = size
            elif size is not None and size != batch_size:
                _LOGGER.warning(
                    "Could not determine batch size unambiguously, found %d and %d",
                    batch_size,
                    size,
                )
                break
    except RecursionError:
        raise RecursionError(error_msg) from None

    if batch_size is None:
        raise RuntimeError(error_msg)

    return batch_size


def sized_len(dataloader) -> int | None:
    """Sized len."""
    try:
        return len(dataloader)
    except (TypeError, NotImplementedError):
        return None



import abc
from collections.abc import Iterator
from typing import TYPE_CHECKING, TypeVar

import beartype
import jaxtyping as jt
from typing_extensions import override

from . import _strategies
from .. import data as data_

if TYPE_CHECKING:
    import reax


_T_co = TypeVar("_T_co", covariant=True)
_U = TypeVar("_U")
_IdxT = TypeVar("_IdxT")


class ParallelStrategy(_strategies.Strategy, abc.ABC):
    @property
    @abc.abstractmethod
    def process_index(self) -> int:
        """Return the global rank of the current process."""

    @property
    @abc.abstractmethod
    def process_count(self) -> int:
        """Return the total number of ranks."""

    @override
    def setup_dataloader(self, data: "reax.DataLoader[_T_co, _U]") -> "reax.DataLoader[_T_co, _U]":
        if not isinstance(data, data_.DataLoader):
            data = super().setup_dataloader(data)

        seed = self.device.id
        sampler = data_.samplers.DistributedSampler(
            dataset=data.dataset,
            num_replicas=self.process_count,
            process_index=self.process_index,
            shuffle=getattr(data, "shuffle", True),
            seed=seed,
            # seed=jnp.array(0, device=self.),
            # drop_last: bool = False,
        )
        # TODO: Maybe try to re-create the BatchSampler with the current sampler
        if isinstance(data.sampler, data_.BatchSampler):
            data.sampler.sampler = sampler
        else:
            data = data.with_new_sampler(sampler)

        return data


class ParallelDataLoader(data_.DataLoader[_T_co, _U]):
    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(
        self,
        dataset: "reax.data.Dataset[_T_co]",
        sampler: data_.samplers.DistributedSampler,
        fetcher: "fetchers._BaseFetcher[_T_co, _U]",
    ):
        self._dataset = dataset
        self._sampler = sampler
        self._fetcher = fetcher

    @override
    @property
    def dataset(self) -> "reax.data.Dataset[_T_co]":
        return self._dataset

    @override
    @property
    def sampler(self):
        return self._sampler

    def __iter__(self) -> Iterator[_U]:
        """Iter function."""
        try:
            for indices in self.sampler:
                yield self._fetcher.fetch(indices)
        except StopIteration:
            pass

    def with_new_sampler(self, sampler: "reax.data.Sampler") -> "ParallelDataLoader[_T_co, _U]":
        return ParallelDataLoader(dataset=self._dataset, sampler=sampler, fetcher=self._fetcher)



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

import beartype
import equinox
import flax.core
import jaxtyping as jt

from . import _metric as metric_

if TYPE_CHECKING:
    import reax

__all__ = ("MetricCollection", "combine")

MetricType = type[metric_.Metric] | metric_.Metric


class MetricCollection(equinox.Module):
    """A collection of metrics that can be created/updated/merged using a single call."""

    _metrics: flax.core.FrozenDict[str, "reax.Metric"]

    @jt.jaxtyped(typechecker=beartype.beartype)
    def __init__(self, metrics: "reax.Metric | Sequence[reax.Metric] | dict[str, reax.Metric]"):
        super().__init__()
        self._metrics = flax.core.FrozenDict(_metrics_dict(metrics))

    def items(self):
        """Items function."""
        return self._metrics.items()

    def empty(self) -> "MetricCollection":
        """Create a new empty instance.

        By default, this will call the constructor with no arguments, if needed, subclasses can
        overwrite this with custom behaviour.
        """
        return MetricCollection({name: metric.empty() for name, metric in self._metrics.items()})

    def create(self, *args, **kwargs) -> "MetricCollection":
        """Create function."""
        return MetricCollection(
            {name: metric.create(*args, **kwargs) for name, metric in self._metrics.items()}
        )

    def update(self, *args, **kwargs) -> "MetricCollection":
        """Update function."""
        return MetricCollection(
            {name: metric.update(*args, **kwargs) for name, metric in self._metrics.items()}
        )

    @jt.jaxtyped(typechecker=beartype.beartype)
    def merge(self, other: "MetricCollection") -> "MetricCollection":
        """Merge function."""
        merged = {}
        for name, metric in self._metrics.items():
            try:
                merged[name] = metric.merge(
                    other._metrics[name]  # pylint: disable=protected-access
                )
            except KeyError:
                merged[name] = metric  # Nothing to merge

        # Now put in all the ones that only appear in 'other'
        for name in set(other._metrics.keys()) - set(  # pylint: disable=protected-access
            self._metrics.keys()
        ):
            merged[name] = other._metrics[name]  # pylint: disable=protected-access

        return MetricCollection(merged)

    def compute(self) -> dict[str, Any]:
        """Compute function."""
        return {name: metric.compute() for name, metric in self._metrics.items()}


def _metrics_dict(
    metrics: "reax.Metric | Sequence[reax.Metric] | dict[str, reax.Metric]",
) -> dict[str, metric_.Metric]:
    """Metrics dict."""
    if isinstance(metrics, dict):
        return {name: _ensure_metric(metric) for name, metric in metrics.items()}

    if isinstance(metrics, Sequence):
        return {type(metric).__name__: metric for metric in map(_ensure_metric, metrics)}

    # Assume we have a single metric
    metric = _ensure_metric(metrics)
    return {type(metric).__name__: metric}


def _ensure_metric(metric: "reax.Metric") -> metric_.Metric:
    """Ensure metric."""
    if not isinstance(metric, metric_.Metric):
        raise TypeError(f"Was expecting a reax.Matric, got: ({type(metric).__name__}) {metric}")

    return metric


def combine(*metric: metric_.Metric) -> MetricCollection:
    """Combine multiple metrics with the same signature into a collection that can be used to
    calculate multiple metrics at once.
    """
    return MetricCollection(metric)
